import asyncio
import copy
from dataclasses import asdict, dataclass
from typing import Any, AsyncIterator, Awaitable, Callable, Dict, List, Optional, Tuple

import structlog
from sanic import Websocket  # type: ignore
from sanic.exceptions import ServerError, WebsocketClosed

from rasa.core.channels import InputChannel, OutputChannel, UserMessage
from rasa.core.channels.voice_ready.utils import (
    CallParameters,
    validate_voice_license_scope,
)
from rasa.core.channels.voice_stream.asr.asr_engine import ASREngine
from rasa.core.channels.voice_stream.asr.asr_event import (
    ASREvent,
    NewTranscript,
    UserIsSpeaking,
)
from rasa.core.channels.voice_stream.asr.azure import AzureASR
from rasa.core.channels.voice_stream.asr.deepgram import DeepgramASR
from rasa.core.channels.voice_stream.audio_bytes import HERTZ, RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import (
    CallState,
    _call_state,
    call_state,
)
from rasa.core.channels.voice_stream.tts.azure import AzureTTS
from rasa.core.channels.voice_stream.tts.cartesia import CartesiaTTS
from rasa.core.channels.voice_stream.tts.tts_cache import TTSCache
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine, TTSError
from rasa.core.channels.voice_stream.util import generate_silence
from rasa.shared.core.constants import SLOT_SILENCE_TIMEOUT
from rasa.shared.utils.cli import print_error_and_exit
from rasa.shared.utils.common import (
    class_from_module_path,
    mark_as_beta_feature,
)
from rasa.utils.io import remove_emojis

logger = structlog.get_logger(__name__)


@dataclass
class VoiceChannelAction:
    pass


@dataclass
class NewAudioAction(VoiceChannelAction):
    audio_bytes: RasaAudioBytes


@dataclass
class EndConversationAction(VoiceChannelAction):
    pass


@dataclass
class ContinueConversationAction(VoiceChannelAction):
    pass


def asr_engine_from_config(asr_config: Dict) -> ASREngine:
    name = str(asr_config["name"])
    asr_config = copy.copy(asr_config)
    asr_config.pop("name")
    if name.lower() == "deepgram":
        return DeepgramASR.from_config_dict(asr_config)
    if name == "azure":
        return AzureASR.from_config_dict(asr_config)
    else:
        mark_as_beta_feature("Custom ASR Engine")
        try:
            asr_engine_class = class_from_module_path(name)
            return asr_engine_class.from_config_dict(asr_config)
        except NameError:
            print_error_and_exit(
                f"Failed to initialize ASR Engine with type '{name}'. "
                f"Please make sure the method `from_config_dict`is implemented."
            )
        except TypeError as e:
            print_error_and_exit(
                f"Failed to initialize ASR Engine with type '{name}'. "
                f"Invalid configuration provided. "
                f"Error: {e}"
            )


def tts_engine_from_config(tts_config: Dict) -> TTSEngine:
    name = str(tts_config["name"])
    tts_config = copy.copy(tts_config)
    tts_config.pop("name")
    if name.lower() == "azure":
        return AzureTTS.from_config_dict(tts_config)
    elif name.lower() == "cartesia":
        return CartesiaTTS.from_config_dict(tts_config)
    else:
        mark_as_beta_feature("Custom TTS Engine")
        try:
            tts_engine_class = class_from_module_path(name)
            return tts_engine_class.from_config_dict(tts_config)
        except NameError:
            print_error_and_exit(
                f"Failed to initialize TTS Engine with type '{name}'. "
                f"Please make sure the method `from_config_dict`is implemented."
            )
        except TypeError as e:
            print_error_and_exit(
                f"Failed to initialize ASR Engine with type '{name}'. "
                f"Invalid configuration provided. "
                f"Error: {e}"
            )


class VoiceOutputChannel(OutputChannel):
    def __init__(
        self,
        voice_websocket: Websocket,
        tts_engine: TTSEngine,
        tts_cache: TTSCache,
    ):
        super().__init__()
        self.voice_websocket = voice_websocket
        self.tts_engine = tts_engine
        self.tts_cache = tts_cache

        self.latest_message_id: Optional[str] = None

    def rasa_audio_bytes_to_channel_bytes(
        self, rasa_audio_bytes: RasaAudioBytes
    ) -> bytes:
        """Turn rasa's audio byte format into the format for the channel."""
        raise NotImplementedError

    def channel_bytes_to_message(self, recipient_id: str, channel_bytes: bytes) -> str:
        """Wrap the bytes for the channel in the proper format."""
        raise NotImplementedError

    def create_marker_message(self, recipient_id: str) -> Tuple[str, str]:
        """Create a marker message for a specific channel."""
        raise NotImplementedError

    async def send_marker_message(self, recipient_id: str) -> None:
        """Send a message that marks positions in the audio stream."""
        marker_message, mark_id = self.create_marker_message(recipient_id)
        await self.voice_websocket.send(marker_message)
        self.latest_message_id = mark_id

    async def send_start_marker(self, recipient_id: str) -> None:
        """Send a marker message before the first audio chunk."""
        # Default implementation uses the generic marker message
        await self.send_marker_message(recipient_id)

    async def send_intermediate_marker(self, recipient_id: str) -> None:
        """Send a marker message during audio streaming."""
        await self.send_marker_message(recipient_id)

    async def send_end_marker(self, recipient_id: str) -> None:
        """Send a marker message after the last audio chunk."""
        await self.send_marker_message(recipient_id)

    def update_silence_timeout(self) -> None:
        """Updates the silence timeout for the session."""
        if self.tracker_state:
            call_state.silence_timeout = (  # type: ignore[attr-defined]
                self.tracker_state["slots"][SLOT_SILENCE_TIMEOUT]
            )

    async def send_text_with_buttons(
        self,
        recipient_id: str,
        text: str,
        buttons: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Uses the concise button output format for voice channels."""
        await self.send_text_with_buttons_concise(recipient_id, text, buttons, **kwargs)

    async def send_text_message(
        self, recipient_id: str, text: str, **kwargs: Any
    ) -> None:
        text = remove_emojis(text)
        self.update_silence_timeout()
        cached_audio_bytes = self.tts_cache.get(text)
        collected_audio_bytes = RasaAudioBytes(b"")
        seconds_marker = -1

        # Send start marker before first chunk
        try:
            await self.send_start_marker(recipient_id)
        except (WebsocketClosed, ServerError):
            call_state.connection_failed = True  # type: ignore[attr-defined]

        if cached_audio_bytes:
            audio_stream = self.chunk_audio(cached_audio_bytes)
        else:
            # Todo: make kwargs compatible with engine config
            synth_config = self.tts_engine.config.__class__.from_dict({})
            try:
                audio_stream = self.tts_engine.synthesize(text, synth_config)
            except TTSError:
                # TODO: add message that works without tts, e.g. loading from disc
                audio_stream = self.chunk_audio(generate_silence())

        async for audio_bytes in audio_stream:
            try:
                await self.send_audio_bytes(recipient_id, audio_bytes)
                full_seconds_of_audio = len(collected_audio_bytes) // HERTZ
                if full_seconds_of_audio > seconds_marker:
                    await self.send_intermediate_marker(recipient_id)
                    seconds_marker = full_seconds_of_audio

            except (WebsocketClosed, ServerError):
                # ignore sending error, and keep collecting and caching audio bytes
                call_state.connection_failed = True  # type: ignore[attr-defined]
            collected_audio_bytes = RasaAudioBytes(collected_audio_bytes + audio_bytes)

        try:
            await self.send_end_marker(recipient_id)
        except (WebsocketClosed, ServerError):
            # ignore sending error
            pass
        call_state.latest_bot_audio_id = self.latest_message_id  # type: ignore[attr-defined]

        if not cached_audio_bytes:
            self.tts_cache.put(text, collected_audio_bytes)

    async def send_audio_bytes(
        self, recipient_id: str, audio_bytes: RasaAudioBytes
    ) -> None:
        channel_bytes = self.rasa_audio_bytes_to_channel_bytes(audio_bytes)
        message = self.channel_bytes_to_message(recipient_id, channel_bytes)
        await self.voice_websocket.send(message)

    async def chunk_audio(
        self, audio_bytes: RasaAudioBytes, chunk_size: int = 2048
    ) -> AsyncIterator[RasaAudioBytes]:
        """Generate chunks from cached audio bytes."""
        offset = 0
        while offset < len(audio_bytes):
            chunk = audio_bytes[offset : offset + chunk_size]
            if len(chunk):
                yield RasaAudioBytes(chunk)
            offset += chunk_size
        return

    async def hangup(self, recipient_id: str, **kwargs: Any) -> None:
        call_state.should_hangup = True  # type: ignore[attr-defined]


class VoiceInputChannel(InputChannel):
    def __init__(
        self,
        server_url: str,
        asr_config: Dict,
        tts_config: Dict,
        monitor_silence: bool = False,
    ):
        validate_voice_license_scope()
        self.server_url = server_url
        self.asr_config = asr_config
        self.tts_config = tts_config
        self.monitor_silence = monitor_silence
        self.tts_cache = TTSCache(tts_config.get("cache_size", 1000))

    async def handle_silence_timeout(
        self,
        voice_websocket: Websocket,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        tts_engine: TTSEngine,
        call_parameters: CallParameters,
    ) -> None:
        timeout = call_state.silence_timeout
        if not timeout:
            return
        if not self.monitor_silence:
            return
        logger.debug("voice_channel.silence_timeout_watch_started", timeout=timeout)
        await asyncio.sleep(timeout)
        logger.debug("voice_channel.silence_timeout_tripped")
        output_channel = self.create_output_channel(voice_websocket, tts_engine)
        message = UserMessage(
            "/silence_timeout",
            output_channel,
            call_parameters.stream_id,
            input_channel=self.name(),
            metadata=asdict(call_parameters),
        )
        await on_new_message(message)

    @staticmethod
    def _cancel_silence_timeout_watcher() -> None:
        """Cancels the silent timeout task if it exists."""
        if call_state.silence_timeout_watcher:
            logger.debug("voice_channel.cancelling_current_timeout_watcher_task")
            call_state.silence_timeout_watcher.cancel()
            call_state.silence_timeout_watcher = None  # type: ignore[attr-defined]

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[str, Any]]) -> InputChannel:
        credentials = credentials or {}
        return cls(
            credentials["server_url"],
            credentials["asr"],
            credentials["tts"],
            credentials.get("monitor_silence", False),
        )

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        raise NotImplementedError

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        raise NotImplementedError

    async def start_session(
        self,
        channel_websocket: Websocket,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        tts_engine: TTSEngine,
        call_parameters: CallParameters,
    ) -> None:
        output_channel = self.create_output_channel(channel_websocket, tts_engine)
        message = UserMessage(
            "/session_start",
            output_channel,
            call_parameters.stream_id,
            input_channel=self.name(),
            metadata=asdict(call_parameters),
        )
        await on_new_message(message)

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        """Map a channel input message to a voice channel action."""
        raise NotImplementedError

    async def run_audio_streaming(
        self,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        channel_websocket: Websocket,
    ) -> None:
        """Pipe input audio to ASR and consume ASR events simultaneously."""
        _call_state.set(CallState())
        asr_engine = asr_engine_from_config(self.asr_config)
        tts_engine = tts_engine_from_config(self.tts_config)
        await asr_engine.connect()

        call_parameters = await self.collect_call_parameters(channel_websocket)
        if call_parameters is None:
            raise ValueError("Failed to extract call parameters for call.")
        await self.start_session(
            channel_websocket, on_new_message, tts_engine, call_parameters
        )

        async def consume_audio_bytes() -> None:
            async for message in channel_websocket:
                is_bot_speaking_before = call_state.is_bot_speaking
                channel_action = self.map_input_message(message, channel_websocket)
                is_bot_speaking_after = call_state.is_bot_speaking

                if not is_bot_speaking_before and is_bot_speaking_after:
                    logger.debug("voice_channel.bot_started_speaking")
                    # relevant when the bot speaks multiple messages in one turn
                    self._cancel_silence_timeout_watcher()

                # we just stopped speaking, starting a watcher for silence timeout
                if is_bot_speaking_before and not is_bot_speaking_after:
                    logger.debug("voice_channel.bot_stopped_speaking")
                    self._cancel_silence_timeout_watcher()
                    call_state.silence_timeout_watcher = (  # type: ignore[attr-defined]
                        asyncio.create_task(
                            self.handle_silence_timeout(
                                channel_websocket,
                                on_new_message,
                                tts_engine,
                                call_parameters,
                            )
                        )
                    )
                if isinstance(channel_action, NewAudioAction):
                    await asr_engine.send_audio_chunks(channel_action.audio_bytes)
                elif isinstance(channel_action, EndConversationAction):
                    # end stream event came from the other side
                    break

        async def consume_asr_events() -> None:
            async for event in asr_engine.stream_asr_events():
                await self.handle_asr_event(
                    event,
                    channel_websocket,
                    on_new_message,
                    tts_engine,
                    call_parameters,
                )

        audio_forwarding_task = asyncio.create_task(consume_audio_bytes())
        asr_event_task = asyncio.create_task(consume_asr_events())
        await asyncio.wait(
            [audio_forwarding_task, asr_event_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if not audio_forwarding_task.done():
            audio_forwarding_task.cancel()
        if not asr_event_task.done():
            asr_event_task.cancel()
        await tts_engine.close_connection()
        await asr_engine.close_connection()
        await channel_websocket.close()
        self._cancel_silence_timeout_watcher()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        """Create a matching voice output channel for this voice input channel."""
        raise NotImplementedError

    async def handle_asr_event(
        self,
        e: ASREvent,
        voice_websocket: Websocket,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        tts_engine: TTSEngine,
        call_parameters: CallParameters,
    ) -> None:
        """Handle a new event from the ASR system."""
        if isinstance(e, NewTranscript) and e.text:
            logger.debug(
                "VoiceInputChannel.handle_asr_event.new_transcript", transcript=e.text
            )
            call_state.is_user_speaking = False  # type: ignore[attr-defined]
            output_channel = self.create_output_channel(voice_websocket, tts_engine)
            message = UserMessage(
                e.text,
                output_channel,
                call_parameters.stream_id,
                input_channel=self.name(),
                metadata=asdict(call_parameters),
            )
            await on_new_message(message)
        elif isinstance(e, UserIsSpeaking):
            self._cancel_silence_timeout_watcher()
            call_state.is_user_speaking = True  # type: ignore[attr-defined]



from __future__ import annotations

from dataclasses import dataclass
from typing import Text

from rasa.shared.core.flows.flow_step_links import FlowStepLinks
from rasa.shared.core.flows.steps.constants import END_STEP, UNSET_FLOW_STEP_ID
from rasa.shared.core.flows.steps.internal import InternalFlowStep


@dataclass
class EndFlowStep(InternalFlowStep):
    """A dynamically added flow step that marks the end of a flow."""

    def __init__(self, flow_id: Text) -> None:
        """Initializes an end flow step."""
        super().__init__(
            idx=UNSET_FLOW_STEP_ID,
            custom_id=END_STEP,
            description=None,
            metadata={},
            next=FlowStepLinks(links=[]),
            flow_id=flow_id,
        )



import argparse
import asyncio
import sys
from typing import Any, Dict, List

import structlog

import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
import rasa.shared.utils.yaml
from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_endpoint_param,
    add_model_param,
    add_remote_storage_param,
)
from rasa.cli.e2e_test import (
    RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME,
    read_test_cases,
    validate_model_path,
)
from rasa.core.exceptions import AgentNotReady
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding.generator import SingleStepLLMCommandGenerator
from rasa.e2e_test.e2e_test_runner import E2ETestRunner
from rasa.llm_fine_tuning.annotation_module import annotate_e2e_tests
from rasa.llm_fine_tuning.llm_data_preparation_module import convert_to_fine_tuning_data
from rasa.llm_fine_tuning.paraphrasing.conversation_rephraser import (
    ConversationRephraser,
)
from rasa.llm_fine_tuning.paraphrasing_module import create_paraphrased_conversations
from rasa.llm_fine_tuning.storage import (
    FileStorageStrategy,
    StorageContext,
    StorageType,
)
from rasa.llm_fine_tuning.train_test_split_module import (
    CONVERSATIONAL_DATA_FORMAT,
    INSTRUCTION_DATA_FORMAT,
    split_llm_fine_tuning_data,
)
from rasa.shared.constants import (
    DEFAULT_ENDPOINTS_PATH,
    DEFAULT_MODELS_PATH,
    LLM_CONFIG_KEY,
)
from rasa.shared.utils.llm import (
    combine_custom_and_default_config,
    resolve_model_client_config,
)
from rasa.shared.utils.yaml import read_config_file
from rasa.utils.beta import ensure_beta_feature_is_enabled

DEFAULT_INPUT_E2E_TEST_PATH = "e2e_tests"
DEFAULT_OUTPUT_FOLDER = "output"
RESULT_SUMMARY_FILE = "result_summary.yaml"
PARAMETERS_FILE = "params.yaml"

structlogger = structlog.get_logger()


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add the llm fine-tuning subparser to `rasa test`.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    llm_parser = subparsers.add_parser(
        "llm",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Commands related to LLMs.",
    )
    llm_subparsers = llm_parser.add_subparsers()

    llm_finetune_parser = llm_subparsers.add_parser(
        "finetune",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Commands related to LLM fine-tuning.",
    )
    llm_finetune_subparser = llm_finetune_parser.add_subparsers()

    create_llm_finetune_data_preparation_subparser(llm_finetune_subparser, parents)


def create_llm_finetune_data_preparation_subparser(
    fine_tune_llm_parser: SubParsersAction,
    parents: List[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    """Create fine-tuning LLM data preparation subparser."""
    data_preparation_subparser = fine_tune_llm_parser.add_parser(
        "prepare-data",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Prepares data for LLM fine-tuning.",
    )

    data_preparation_subparser.set_defaults(func=prepare_llm_fine_tuning_data)

    add_data_preparation_arguments(data_preparation_subparser)
    add_model_param(data_preparation_subparser, add_positional_arg=False)
    add_endpoint_param(
        data_preparation_subparser,
        help_text="Configuration file for the model server and the connectors as a "
        "yml file.",
    )

    return data_preparation_subparser


def add_data_preparation_arguments(parser: argparse.ArgumentParser) -> None:
    """Arguments for preparing LLM fine-tuning data."""
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        default=DEFAULT_OUTPUT_FOLDER,
        help="The output folder to store the data to.",
    )
    parser.add_argument(
        "path-to-e2e-test-cases",
        nargs="?",
        type=str,
        help="Input file or folder containing end-to-end test cases.",
        default=DEFAULT_INPUT_E2E_TEST_PATH,
    )

    add_remote_storage_param(parser)

    rephrasing_arguments = parser.add_argument_group("Rephrasing Module")
    rephrasing_arguments.add_argument(
        "--num-rephrases",
        choices=range(0, 50),
        type=int,
        default=10,
        help="Number of rephrases to be generated per user utterance.",
    )
    rephrasing_arguments.add_argument(
        "--rephrase-config",
        type=str,
        default=None,
        help="Path to config file that contains the configuration of the "
        "rephrasing module.",
    )

    train_test_split_arguments = parser.add_argument_group("Train/Test Split Module")
    train_test_split_arguments.add_argument(
        "--train-frac",
        type=restricted_float,
        default=0.8,
        help="The amount of data that should go into the training dataset. The value "
        "should be >0.0 and <=1.0.",
    )
    train_test_split_arguments.add_argument(
        "--output-format",
        choices=[INSTRUCTION_DATA_FORMAT, CONVERSATIONAL_DATA_FORMAT],
        type=str,
        nargs="?",
        default=INSTRUCTION_DATA_FORMAT,
        help="Format of the output file.",
    )


def prepare_llm_fine_tuning_data(args: argparse.Namespace) -> None:
    """Prepare LLM fine-tuning data.

    Args:
        args: Commandline arguments.
    """
    ensure_beta_feature_is_enabled(
        "LLM fine-tuning recipe",
        env_flag=RASA_PRO_BETA_FINE_TUNING_RECIPE_ENV_VAR_NAME,
    )

    rephrase_config = (
        read_config_file(args.rephrase_config) if args.rephrase_config else {}
    )
    ConversationRephraser.validate_config(rephrase_config)

    # make sure the output directory exists
    output_dir = args.out
    rasa.shared.utils.io.create_directory(output_dir)

    # read e2e test cases
    path_to_test_cases = getattr(
        args, "path-to-e2e-test-cases", DEFAULT_INPUT_E2E_TEST_PATH
    )
    test_suite = read_test_cases(path_to_test_cases)
    # set up the e2e test runner
    e2e_test_runner = set_up_e2e_test_runner(args)

    if e2e_test_runner.agent.processor is None:
        rasa.shared.utils.cli.print_error(
            "No processor: Not able to retrieve flows and config from trained model."
        )
        sys.exit(0)

    flows = asyncio.run(e2e_test_runner.agent.processor.get_flows())
    llm_command_generator_config = _get_llm_command_generator_config(e2e_test_runner)

    # set up storage context
    storage_context = create_storage_context(StorageType.FILE, output_dir)

    statistics = {}

    # 1. annotate e2e tests
    log_start_of_module("Annotation")
    conversations = annotate_e2e_tests(e2e_test_runner, test_suite, storage_context)
    statistics["num_input_e2e_tests"] = len(test_suite.test_cases)
    statistics["num_annotated_conversations"] = len(conversations)
    statistics["num_user_messages_across_conversations"] = sum(
        [len(conversation.get_user_messages()) for conversation in conversations]
    )
    statistics["num_user_messages_to_rephrase_across_conversations"] = sum(
        [
            len(conversation.get_user_messages_to_rephrase())
            for conversation in conversations
        ]
    )
    log_end_of_module("Annotation", statistics)

    # 2. paraphrase conversations
    log_start_of_module("Rephrasing")
    conversations, rephrase_config = asyncio.run(
        create_paraphrased_conversations(
            conversations,
            rephrase_config,
            args.num_rephrases,
            flows,
            llm_command_generator_config,
            storage_context,
        )
    )
    statistics["num_passing_rephrased_user_messages"] = sum(
        [conversation.get_number_of_rephrases(True) for conversation in conversations]
    )
    statistics["num_failing_rephrased_user_messages"] = sum(
        [conversation.get_number_of_rephrases(False) for conversation in conversations]
    )
    log_end_of_module("Rephrasing", statistics)

    # 3. create fine-tuning dataset
    log_start_of_module("LLM Data Preparation")
    llm_fine_tuning_data = convert_to_fine_tuning_data(conversations, storage_context)
    statistics["num_ft_data_points"] = len(llm_fine_tuning_data)
    log_end_of_module("LLM Data Preparation", statistics)

    # 4. create train/test split
    log_start_of_module("Train/Test Split")
    train_data, val_data = split_llm_fine_tuning_data(
        llm_fine_tuning_data,
        args.train_frac,
        args.output_format,
        storage_context,
        test_suite,
    )
    statistics["num_train_data_points"] = len(train_data)
    statistics["num_val_data_points"] = len(val_data)
    log_end_of_module("Train/Test Split", statistics)

    # write down params and statistics to a file
    write_params(args, rephrase_config, output_dir)
    write_statistics(statistics, output_dir)

    rasa.shared.utils.cli.print_success(
        f"Data and intermediate results are written " f"to '{output_dir}'."
    )


def _get_llm_command_generator_config(e2e_test_runner: E2ETestRunner) -> Dict[str, Any]:
    from rasa.dialogue_understanding.generator.constants import DEFAULT_LLM_CONFIG

    train_schema = e2e_test_runner.agent.processor.model_metadata.train_schema  # type: ignore

    for node_name, node in train_schema.nodes.items():
        if node.matches_type(SingleStepLLMCommandGenerator, include_subtypes=True):
            # Configurations can reference model groups defined in the endpoints.yml
            resolved_config = resolve_model_client_config(
                node.config.get(LLM_CONFIG_KEY, {}), node_name
            )
            return combine_custom_and_default_config(
                resolved_config, DEFAULT_LLM_CONFIG
            )

    rasa.shared.utils.cli.print_error(
        "The provided model is not trained using 'SingleStepLLMCommandGenerator' or "
        "its subclasses. Without it, no data for fine-tuning can be generated. To "
        "resolve this, please include 'SingleStepLLMCommandGenerator' or its subclass "
        "in your config and train your model."
    )
    sys.exit(1)


def log_start_of_module(module_name: str) -> None:
    log_info = f"Starting {module_name} Module"
    rasa.shared.utils.cli.print_info(
        f"{rasa.shared.utils.cli.pad(log_info, char='-')}\n"
    )


def log_end_of_module(module_name: str, statistics: Dict[str, int]) -> None:
    log_info = f"Finished {module_name} Module"
    rasa.shared.utils.cli.print_info(
        f"{rasa.shared.utils.cli.pad(log_info, char='-')}\n"
    )
    rasa.shared.utils.cli.print_color(
        "Current Statistics:", color=rasa.shared.utils.io.bcolors.BOLD
    )
    for key, value in statistics.items():
        rasa.shared.utils.cli.print_color(
            f"  {key}: {value}", color=rasa.shared.utils.io.bcolors.BOLD
        )


def restricted_float(x: Any) -> float:
    try:
        x = float(x)
    except ValueError:
        raise argparse.ArgumentTypeError("%r not a floating-point literal" % (x,))

    if x <= 0.0 or x > 1.0:
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]" % (x,))
    return x


def write_params(
    args: argparse.Namespace, rephrase_config: Dict[str, Any], output_path: str
) -> None:
    yaml_data = {
        "parameters": {
            "num_rephrases": args.num_rephrases,
            "rephrase_config": rephrase_config,
            "model": args.model,
            "endpoints": args.endpoints,
            "remote-storage": args.remote_storage,
            "train_frac": args.train_frac,
            "output_format": args.output_format,
            "out": output_path,
        }
    }

    rasa.shared.utils.yaml.write_yaml(yaml_data, f"{output_path}/{PARAMETERS_FILE}")


def write_statistics(statistics: Dict[str, Any], output_path: str) -> None:
    rasa.shared.utils.yaml.write_yaml(
        statistics, f"{output_path}/{RESULT_SUMMARY_FILE}"
    )


def get_valid_endpoints(endpoints_file: str) -> AvailableEndpoints:
    validated_endpoints_file = rasa.cli.utils.get_validated_path(
        endpoints_file, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )
    endpoints = AvailableEndpoints.get_instance(validated_endpoints_file)

    # Ignore all endpoints apart from action server, model, nlu and nlg
    # to ensure InMemoryTrackerStore is being used instead of production
    # tracker store
    endpoints.tracker_store = None
    endpoints.lock_store = None
    endpoints.event_broker = None

    return endpoints


def set_up_e2e_test_runner(args: argparse.Namespace) -> E2ETestRunner:
    endpoints = get_valid_endpoints(args.endpoints)

    if endpoints.model is None:
        args.model = validate_model_path(args.model, "model", DEFAULT_MODELS_PATH)

    try:
        return E2ETestRunner(
            remote_storage=args.remote_storage,
            model_path=args.model,
            model_server=endpoints.model,
            endpoints=endpoints,
        )
    except AgentNotReady as error:
        structlogger.error(
            "cli.finetune_llm.prepare_data.set_up_e2e_test_runner", error=error.message
        )
        sys.exit(1)


def create_storage_context(
    storage_type: StorageType, output_dir: str
) -> StorageContext:
    if storage_type == StorageType.FILE:
        strategy = FileStorageStrategy(output_dir)
    else:
        raise ValueError("Unsupported storage type")

    return StorageContext(strategy)



from dataclasses import dataclass

from rasa.dialogue_understanding.stack.frames import BaseFlowStackFrame


@dataclass
class PatternFlowStackFrame(BaseFlowStackFrame):
    """A stack frame that represents a pattern flow."""

    pass



import base64
import os
from typing import Optional

import structlog

from rasa.model_manager import config
from rasa.shared.exceptions import RasaException

structlogger = structlog.get_logger()


class InvalidPathException(RasaException):
    """Raised if a path is invalid - e.g. path traversal is detected."""


def write_encoded_data_to_file(encoded_data: bytes, file: str) -> None:
    """Write base64 encoded data to a file."""
    # create the directory if it does not exist of the parent directory
    os.makedirs(os.path.dirname(file), exist_ok=True)

    with open(file, "w") as f:
        decoded = base64.b64decode(encoded_data)
        text = decoded.decode("utf-8")
        f.write(text)


def logs_base_path() -> str:
    """Return the path to the logs' directory."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY, "logs")


def ensure_base_directory_exists(directory: str) -> None:
    """Ensure that a files parent directory exists.

    Args:
        directory: The directory to check.
    """
    os.makedirs(os.path.dirname(directory), exist_ok=True)


def models_base_path() -> str:
    """Return the path to the models' directory."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY, "models")


def logs_path(action_id: str) -> str:
    """Return the path to the log file for a given action id.

    Args:
        action_id: can either be a training_id or a deployment_id
    """
    return subpath(logs_base_path(), f"{action_id}.txt")


def subpath(parent: str, child: str) -> str:
    """Return the path to the child directory of the parent directory.

    Ensures, that child doesn't navigate to parent directories. Prevents
    path traversal. Raises an InvalidPathException if the path is invalid.

    Based on Snyk's directory traversal mitigation:
    https://learn.snyk.io/lesson/directory-traversal/
    """
    safe_path = os.path.abspath(os.path.join(parent, child))
    parent = os.path.abspath(parent)

    common_base = os.path.commonpath([parent, safe_path])
    if common_base != parent:
        raise InvalidPathException(f"Invalid path: {safe_path}")

    if os.path.basename(safe_path) != child:
        raise InvalidPathException(
            f"Invalid path - path traversal detected: {safe_path}"
        )

    return safe_path


def get_logs_content(action_id: str) -> Optional[str]:
    """Return the content of the log file for a given action id."""
    try:
        with open(logs_path(action_id), "r") as file:
            return file.read()
    except FileNotFoundError:
        structlogger.debug("model_service.logs.not_found", action_id=action_id)
        return None



DEFAULT_ENCODING = "utf-8"

READ_YAML_FILE_CACHE_MAXSIZE_ENV_VAR = "READ_YAML_FILE_CACHE_MAXSIZE"
DEFAULT_READ_YAML_FILE_CACHE_MAXSIZE = 256
RASA_PRO_BETA_PREDICATES_IN_RESPONSE_CONDITIONS_ENV_VAR_NAME = (
    "RASA_PRO_BETA_PREDICATES_IN_RESPONSE_CONDITIONS"
)



import importlib.resources
import json
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text

import dotenv
import structlog
from jinja2 import Template
from pydantic import ValidationError

import rasa.shared.utils.io
from rasa.core.constants import (
    POLICY_MAX_HISTORY,
    POLICY_PRIORITY,
    SEARCH_POLICY_PRIORITY,
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.core.information_retrieval import (
    InformationRetrieval,
    InformationRetrievalException,
    SearchResult,
    create_from_endpoint_config,
)
from rasa.core.information_retrieval.faiss import FAISS_Store
from rasa.core.policies.policy import Policy, PolicyPrediction
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding.generator.constants import (
    LLM_CONFIG_KEY,
)
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames import (
    DialogueStackFrame,
    PatternFlowStackFrame,
    SearchStackFrame,
)
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.graph_components.providers.forms_provider import Forms
from rasa.graph_components.providers.responses_provider import Responses
from rasa.shared.constants import (
    EMBEDDINGS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_GROUP_ID_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROMPT_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.core.constants import (
    ACTION_CANCEL_FLOW,
    ACTION_SEND_TEXT_NAME,
    DEFAULT_SLOT_NAMES,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import BotUttered, Event, UserUttered
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.shared.exceptions import FileIOException, RasaException
from rasa.shared.nlu.constants import (
    KEY_COMPONENT_NAME,
    KEY_LLM_RESPONSE_METADATA,
    KEY_PROMPT_NAME,
    KEY_USER_PROMPT,
    PROMPTS,
)
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.providers.embedding._langchain_embedding_client_adapter import (
    _LangchainEmbeddingClientAdapter,
)
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.providers.llm.llm_response import LLMResponse, measure_llm_latency
from rasa.shared.utils.cli import print_error_and_exit
from rasa.shared.utils.health_check.embeddings_health_check_mixin import (
    EmbeddingsHealthCheckMixin,
)
from rasa.shared.utils.health_check.llm_health_check_mixin import LLMHealthCheckMixin
from rasa.shared.utils.io import deep_container_fingerprint
from rasa.shared.utils.llm import (
    DEFAULT_OPENAI_CHAT_MODEL_NAME,
    DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
    embedder_factory,
    get_prompt_template,
    llm_factory,
    resolve_model_client_config,
    sanitize_message_for_prompt,
    tracker_as_readable_transcript,
)
from rasa.telemetry import (
    track_enterprise_search_policy_predict,
    track_enterprise_search_policy_train_completed,
    track_enterprise_search_policy_train_started,
)

if TYPE_CHECKING:
    from langchain.schema.embeddings import Embeddings

    from rasa.core.featurizers.tracker_featurizers import TrackerFeaturizer

from rasa.utils.log_utils import log_llm

logger = structlog.get_logger()

dotenv.load_dotenv("./.env")

SOURCE_PROPERTY = "source"
VECTOR_STORE_TYPE_PROPERTY = "type"
VECTOR_STORE_PROPERTY = "vector_store"
VECTOR_STORE_THRESHOLD_PROPERTY = "threshold"
TRACE_TOKENS_PROPERTY = "trace_prompt_tokens"
CITATION_ENABLED_PROPERTY = "citation_enabled"
USE_LLM_PROPERTY = "use_generative_llm"
MAX_MESSAGES_IN_QUERY_KEY = "max_messages_in_query"

DEFAULT_VECTOR_STORE_TYPE = "faiss"
DEFAULT_VECTOR_STORE_THRESHOLD = 0.0
DEFAULT_VECTOR_STORE = {
    VECTOR_STORE_TYPE_PROPERTY: DEFAULT_VECTOR_STORE_TYPE,
    SOURCE_PROPERTY: "./docs",
    VECTOR_STORE_THRESHOLD_PROPERTY: DEFAULT_VECTOR_STORE_THRESHOLD,
}

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: DEFAULT_OPENAI_CHAT_MODEL_NAME,
    TIMEOUT_CONFIG_KEY: 10,
    "temperature": 0.0,
    "max_tokens": 256,
    "max_retries": 1,
}

DEFAULT_EMBEDDINGS_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    "model": DEFAULT_OPENAI_EMBEDDING_MODEL_NAME,
}

ENTERPRISE_SEARCH_PROMPT_FILE_NAME = "enterprise_search_policy_prompt.jinja2"
ENTERPRISE_SEARCH_CONFIG_FILE_NAME = "config.json"

SEARCH_RESULTS_METADATA_KEY = "search_results"
SEARCH_QUERY_METADATA_KEY = "search_query"

DEFAULT_ENTERPRISE_SEARCH_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.core.policies", "enterprise_search_prompt_template.jinja2"
)

DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_CITATION_TEMPLATE = importlib.resources.read_text(
    "rasa.core.policies", "enterprise_search_prompt_with_citation_template.jinja2"
)


class VectorStoreConnectionError(RasaException):
    """Exception raised for errors in connecting to the vector store."""


class VectorStoreConfigurationError(RasaException):
    """Exception raised for errors in vector store configuration."""


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITH_END_TO_END_SUPPORT, is_trainable=True
)
class EnterpriseSearchPolicy(LLMHealthCheckMixin, EmbeddingsHealthCheckMixin, Policy):
    """Policy which uses a vector store and LLMs to respond to user messages.

    The policy uses a vector store and LLMs to respond to user messages. The
    vector store is used to retrieve the most relevant responses to the user
    message. The LLMs are used to rank the responses and select the best
    response. The policy can be used to respond to user messages without
    training data.

    Example Configuration:

        policies:
            # - ...
            - name: EnterpriseSearchPolicy
              vector_store:
                type: "milvus"
                <vector_store_config>
            # - ...
    """

    @staticmethod
    def does_support_stack_frame(frame: DialogueStackFrame) -> bool:
        """Checks if the policy supports the given stack frame."""
        return isinstance(frame, SearchStackFrame)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """Returns the default config of the policy."""
        return {
            POLICY_PRIORITY: SEARCH_POLICY_PRIORITY,
            VECTOR_STORE_PROPERTY: DEFAULT_VECTOR_STORE,
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        vector_store: Optional[InformationRetrieval] = None,
        featurizer: Optional["TrackerFeaturizer"] = None,
        prompt_template: Optional[Text] = None,
    ) -> None:
        """Constructs a new Policy object."""
        super().__init__(config, model_storage, resource, execution_context, featurizer)

        # Resolve LLM config
        self.config[LLM_CONFIG_KEY] = resolve_model_client_config(
            self.config.get(LLM_CONFIG_KEY), EnterpriseSearchPolicy.__name__
        )
        # Resolve embeddings config
        self.config[EMBEDDINGS_CONFIG_KEY] = resolve_model_client_config(
            self.config.get(EMBEDDINGS_CONFIG_KEY), EnterpriseSearchPolicy.__name__
        )

        # Vector store object and configuration
        self.vector_store = vector_store
        self.vector_store_config = self.config.get(
            VECTOR_STORE_PROPERTY, DEFAULT_VECTOR_STORE
        )

        # Embeddings configuration for encoding the search query
        self.embeddings_config = (
            self.config[EMBEDDINGS_CONFIG_KEY] or DEFAULT_EMBEDDINGS_CONFIG
        )

        # LLM Configuration for response generation
        self.llm_config = self.config[LLM_CONFIG_KEY] or DEFAULT_LLM_CONFIG

        # Maximum number of turns to include in the prompt
        self.max_history = self.config.get(POLICY_MAX_HISTORY)

        # Maximum number of messages to include in the search query
        self.max_messages_in_query = self.config.get(MAX_MESSAGES_IN_QUERY_KEY, 2)

        # boolean to enable/disable tracing of prompt tokens
        self.trace_prompt_tokens = self.config.get(TRACE_TOKENS_PROPERTY, False)

        # boolean to enable/disable the use of LLM for response generation
        self.use_llm = self.config.get(USE_LLM_PROPERTY, True)

        # boolean to enable/disable citation generation
        self.citation_enabled = self.config.get(CITATION_ENABLED_PROPERTY, False)

        self.prompt_template = prompt_template or get_prompt_template(
            self.config.get(PROMPT_CONFIG_KEY),
            DEFAULT_ENTERPRISE_SEARCH_PROMPT_TEMPLATE,
        )
        self.citation_prompt_template = get_prompt_template(
            self.config.get(PROMPT_CONFIG_KEY),
            DEFAULT_ENTERPRISE_SEARCH_PROMPT_WITH_CITATION_TEMPLATE,
        )
        # If citation is enabled, use the citation prompt template
        if self.citation_enabled:
            self.prompt_template = self.citation_prompt_template

    @classmethod
    def _create_plain_embedder(cls, config: Dict[Text, Any]) -> "Embeddings":
        """Creates an embedder based on the given configuration.

        Returns:
        The embedder.
        """
        # Copy the config so original config is not modified
        config = config.copy()
        # Resolve config and instantiate the embedding client
        config[EMBEDDINGS_CONFIG_KEY] = resolve_model_client_config(
            config.get(EMBEDDINGS_CONFIG_KEY), EnterpriseSearchPolicy.__name__
        )
        client = embedder_factory(
            config.get(EMBEDDINGS_CONFIG_KEY), DEFAULT_EMBEDDINGS_CONFIG
        )
        # Wrap the embedding client in the adapter
        return _LangchainEmbeddingClientAdapter(client)

    @classmethod
    def _add_prompt_and_llm_response_to_latest_message(
        cls,
        tracker: DialogueStateTracker,
        prompt_name: str,
        user_prompt: str,
        llm_response: Optional[LLMResponse] = None,
    ) -> None:
        """Stores the prompt and LLMResponse metadata in the tracker.

        Args:
            tracker: The DialogueStateTracker containing the current conversation state.
            prompt_name: A name identifying prompt usage.
            user_prompt: The user prompt that was sent to the LLM.
            llm_response: The response object from the LLM (None if no response).
        """
        from rasa.dialogue_understanding.utils import record_commands_and_prompts

        if not record_commands_and_prompts:
            return

        if not tracker.latest_message:
            return

        parse_data = tracker.latest_message.parse_data
        if PROMPTS not in parse_data:
            parse_data[PROMPTS] = []  # type: ignore[literal-required]

        prompt_data: Dict[Text, Any] = {
            KEY_COMPONENT_NAME: cls.__name__,
            KEY_PROMPT_NAME: prompt_name,
            KEY_USER_PROMPT: user_prompt,
            KEY_LLM_RESPONSE_METADATA: llm_response.to_dict() if llm_response else None,
        }

        parse_data[PROMPTS].append(prompt_data)  # type: ignore[literal-required]

    def train(  # type: ignore[override]
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        responses: Responses,
        forms: Forms,
        training_data: TrainingData,
        **kwargs: Any,
    ) -> Resource:
        """Trains a policy.

        Args:
            training_trackers: The story and rules trackers from the training data.
            domain: The model's domain.
            responses: The model's responses.
            forms: The model's forms.
            training_data: The model's training data.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to train itself.

        Returns:
            A policy must return its resource locator so that potential children nodes
            can load the policy from the resource.
        """
        # Perform health checks for both LLM and embeddings client configs
        self._perform_health_checks(self.config, "enterprise_search_policy.train")

        store_type = self.vector_store_config.get(VECTOR_STORE_TYPE_PROPERTY)

        # telemetry call to track training start
        track_enterprise_search_policy_train_started()

        # validate embedding configuration
        try:
            embeddings = self._create_plain_embedder(self.config)
        except (ValidationError, Exception) as e:
            logger.error(
                "enterprise_search_policy.train.embedder_instantiation_failed",
                message="Unable to instantiate the embedding client.",
                error=e,
            )
            print_error_and_exit(
                "Unable to create embedder. Please make sure you specified the "
                f"required environment variables. Error: {e}"
            )

        if store_type == DEFAULT_VECTOR_STORE_TYPE:
            logger.info("enterprise_search_policy.train.faiss")
            with self._model_storage.write_to(self._resource) as path:
                self.vector_store = FAISS_Store(
                    docs_folder=self.vector_store_config.get(SOURCE_PROPERTY),
                    embeddings=embeddings,
                    index_path=path,
                    create_index=True,
                )
        else:
            logger.info("enterprise_search_policy.train.custom", store_type=store_type)

        # telemetry call to track training completion
        track_enterprise_search_policy_train_completed(
            vector_store_type=store_type,
            embeddings_type=self.embeddings_config.get(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_config.get(MODEL_CONFIG_KEY)
            or self.embeddings_config.get(MODEL_NAME_CONFIG_KEY),
            embeddings_model_group_id=self.embeddings_config.get(
                MODEL_GROUP_ID_CONFIG_KEY
            ),
            llm_type=self.llm_config.get(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_config.get(MODEL_CONFIG_KEY)
            or self.llm_config.get(MODEL_NAME_CONFIG_KEY),
            llm_model_group_id=self.llm_config.get(MODEL_GROUP_ID_CONFIG_KEY),
            citation_enabled=self.citation_enabled,
        )
        self.persist()
        return self._resource

    def persist(self) -> None:
        """Persists the policy to storage."""
        with self._model_storage.write_to(self._resource) as path:
            rasa.shared.utils.io.write_text_file(
                self.prompt_template, path / ENTERPRISE_SEARCH_PROMPT_FILE_NAME
            )
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                path / ENTERPRISE_SEARCH_CONFIG_FILE_NAME, self.config
            )

    def _prepare_slots_for_template(
        self, tracker: DialogueStateTracker
    ) -> List[Dict[str, str]]:
        """Prepares the slots for the template.

        Args:
            tracker: The tracker containing the conversation history up to now.

        Returns:
            The non-empty slots.
        """
        template_slots = []
        for name, slot in tracker.slots.items():
            if name not in DEFAULT_SLOT_NAMES and slot.value is not None:
                template_slots.append(
                    {
                        "name": name,
                        "value": str(slot.value),
                        "type": slot.type_name,
                    }
                )
        return template_slots

    def _connect_vector_store_or_raise(
        self, endpoints: Optional[AvailableEndpoints]
    ) -> None:
        """Connects to the vector store or raises an exception.

        Raise exceptions for the following cases:
        - The configuration is not specified
        - Unable to connect to the vector store

        Args:
            endpoints: Endpoints configuration.
        """
        config = endpoints.vector_store if endpoints else None
        store_type = self.vector_store_config.get(VECTOR_STORE_TYPE_PROPERTY)
        if config is None and store_type != DEFAULT_VECTOR_STORE_TYPE:
            logger.error(
                "enterprise_search_policy._connect_vector_store_or_raise.no_config"
            )
            raise VectorStoreConfigurationError(
                """No vector store specified. Please specify a vector
                store in the endpoints configuration"""
            )
        try:
            self.vector_store.connect(config)  # type: ignore
        except Exception as e:
            logger.error(
                "enterprise_search_policy._connect_vector_store_or_raise.connect_error",
                error=e,
                config=config,
            )
            raise VectorStoreConnectionError(
                f"Unable to connect to the vector store. Error: {e}"
            )

    def _prepare_search_query(self, tracker: DialogueStateTracker, history: int) -> str:
        """Prepares the search query.
        The search query is the last N messages in the conversation history.

        Args:
            tracker: The tracker containing the conversation history up to now.
            history: The number of messages to include in the search query.

        Returns:
            The search query.
        """
        transcript = []
        for event in tracker.applied_events():
            if isinstance(event, UserUttered) or isinstance(event, BotUttered):
                transcript.append(sanitize_message_for_prompt(event.text))

        search_query = " ".join(transcript[-history:][::-1])
        logger.debug("search_query", search_query=search_query)
        return search_query

    async def predict_action_probabilities(  # type: ignore[override]
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        endpoints: Optional[AvailableEndpoints],
        rule_only_data: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action the bot should take after seeing the tracker.

        Args:
            tracker: The tracker containing the conversation history up to now.
            domain: The model's domain.
            endpoints: The model's endpoints.
            rule_only_data: Slots and loops which are specific to rules and hence
                should be ignored by this policy.
            **kwargs: Depending on the specified `needs` section and the resulting
                graph structure the policy can use different input to make predictions.

        Returns:
             The prediction.
        """
        logger_key = "enterprise_search_policy.predict_action_probabilities"
        vector_search_threshold = self.vector_store_config.get(
            VECTOR_STORE_THRESHOLD_PROPERTY, DEFAULT_VECTOR_STORE_THRESHOLD
        )
        llm = llm_factory(self.config.get(LLM_CONFIG_KEY), DEFAULT_LLM_CONFIG)
        if not self.supports_current_stack_frame(
            tracker, False, False
        ) or self.should_abstain_in_coexistence(tracker, True):
            return self._prediction(self._default_predictions(domain))

        if not self.vector_store:
            logger.error(f"{logger_key}.no_vector_store")
            return self._create_prediction_internal_error(domain, tracker)

        try:
            self._connect_vector_store_or_raise(endpoints)
        except (VectorStoreConfigurationError, VectorStoreConnectionError) as e:
            logger.error(f"{logger_key}.connection_error", error=e)
            return self._create_prediction_internal_error(domain, tracker)

        search_query = self._prepare_search_query(
            tracker, int(self.max_messages_in_query)
        )
        tracker_state = tracker.current_state(EventVerbosity.AFTER_RESTART)

        try:
            documents = await self.vector_store.search(
                query=search_query,
                tracker_state=tracker_state,
                threshold=vector_search_threshold,
            )
        except InformationRetrievalException as e:
            logger.error(f"{logger_key}.search_error", error=e)
            return self._create_prediction_internal_error(domain, tracker)

        if not documents.results:
            logger.info(f"{logger_key}.no_documents")
            return self._create_prediction_cannot_handle(domain, tracker)

        if self.use_llm:
            prompt = self._render_prompt(tracker, documents.results)
            llm_response = await self._generate_llm_answer(llm, prompt)
            llm_response = LLMResponse.ensure_llm_response(llm_response)

            self._add_prompt_and_llm_response_to_latest_message(
                tracker=tracker,
                prompt_name="enterprise_search_prompt",
                user_prompt=prompt,
                llm_response=llm_response,
            )

            if llm_response is None or not llm_response.choices:
                logger.debug(f"{logger_key}.no_llm_response")
                response = None
            else:
                llm_answer = llm_response.choices[0]

                if self.citation_enabled:
                    llm_answer = self.post_process_citations(llm_answer)

                logger.debug(f"{logger_key}.llm_answer", llm_answer=llm_answer)
                response = llm_answer
        else:
            response = documents.results[0].metadata.get("answer", None)
            if not response:
                logger.error(
                    f"{logger_key}.answer_key_missing_in_metadata",
                    search_results=documents.results,
                )
            logger.debug(
                "enterprise_search_policy.predict_action_probabilities.no_llm",
                search_results=documents,
            )
        if response is None:
            return self._create_prediction_internal_error(domain, tracker)

        action_metadata = {
            "message": {
                "text": response,
                SEARCH_RESULTS_METADATA_KEY: [
                    result.text for result in documents.results
                ],
                UTTER_SOURCE_METADATA_KEY: self.__class__.__name__,
                SEARCH_QUERY_METADATA_KEY: search_query,
            }
        }

        # telemetry call to track policy prediction
        track_enterprise_search_policy_predict(
            vector_store_type=self.vector_store_config.get(VECTOR_STORE_TYPE_PROPERTY),
            embeddings_type=self.embeddings_config.get(PROVIDER_CONFIG_KEY),
            embeddings_model=self.embeddings_config.get(MODEL_CONFIG_KEY)
            or self.embeddings_config.get(MODEL_NAME_CONFIG_KEY),
            embeddings_model_group_id=self.embeddings_config.get(
                MODEL_GROUP_ID_CONFIG_KEY
            ),
            llm_type=self.llm_config.get(PROVIDER_CONFIG_KEY),
            llm_model=self.llm_config.get(MODEL_CONFIG_KEY)
            or self.llm_config.get(MODEL_NAME_CONFIG_KEY),
            llm_model_group_id=self.llm_config.get(MODEL_GROUP_ID_CONFIG_KEY),
            citation_enabled=self.citation_enabled,
        )
        return self._create_prediction(
            domain=domain, tracker=tracker, action_metadata=action_metadata
        )

    def _render_prompt(
        self, tracker: DialogueStateTracker, documents: List[SearchResult]
    ) -> Text:
        """Renders the prompt from the template.

        Args:
            tracker: The tracker containing the conversation history up to now.
            documents: The documents retrieved from search

        Returns:
            The rendered prompt.
        """
        inputs = {
            "current_conversation": tracker_as_readable_transcript(
                tracker, max_turns=self.max_history
            ),
            "docs": documents,
            "slots": self._prepare_slots_for_template(tracker),
            "citation_enabled": self.citation_enabled,
        }
        prompt = Template(self.prompt_template).render(**inputs)
        log_llm(
            logger=logger,
            log_module="EnterpriseSearchPolicy",
            log_event="enterprise_search_policy._render_prompt.prompt_rendered",
            prompt=prompt,
        )
        return prompt

    @measure_llm_latency
    async def _generate_llm_answer(
        self, llm: LLMClient, prompt: Text
    ) -> Optional[LLMResponse]:
        """Fetches an LLM completion for the provided prompt.

        Args:
            llm: The LLM client used to get the completion.
            prompt: The prompt text to send to the model.

        Returns:
            An LLMResponse object, or None if the call fails.
        """
        try:
            return await llm.acompletion(prompt)
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            logger.error(
                "enterprise_search_policy._generate_llm_answer.llm_error",
                error=e,
            )
            return None

    def _create_prediction(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        action_metadata: Dict[Text, Any],
    ) -> PolicyPrediction:
        """Create a policy prediction result with ACTION_SEND_TEXT_NAME.

        Args:
            domain: The model's domain.
            tracker: The tracker containing the conversation history up to now.
            action_metadata: The metadata for the predicted action.

        Returns:
            The prediction.
        """
        result = self._prediction_result(ACTION_SEND_TEXT_NAME, domain)
        stack = tracker.stack
        if not stack.is_empty():
            stack.pop()
            events: List[Event] = tracker.create_stack_updated_events(stack)
        else:
            events = []

        return self._prediction(result, action_metadata=action_metadata, events=events)

    def _create_prediction_internal_error(
        self, domain: Domain, tracker: DialogueStateTracker
    ) -> PolicyPrediction:
        return self._create_prediction_for_pattern(
            domain, tracker, InternalErrorPatternFlowStackFrame()
        )

    def _create_prediction_cannot_handle(
        self, domain: Domain, tracker: DialogueStateTracker
    ) -> PolicyPrediction:
        return self._create_prediction_for_pattern(
            domain, tracker, CannotHandlePatternFlowStackFrame()
        )

    def _create_prediction_for_pattern(
        self,
        domain: Domain,
        tracker: DialogueStateTracker,
        pattern_stack_frame: PatternFlowStackFrame,
    ) -> PolicyPrediction:
        """Create a policy prediction result for error.

        We should cancel the current flow (hence ACTION_CANCEL_FLOW) and push a
        pattern stack frame (Internal Error Pattern by default) to start the pattern.

        Args:
            domain: The model's domain.
            tracker: The tracker containing the conversation history up to now.
            pattern_stack_frame: The pattern stack frame to push.

        Returns:
            The prediction.
        """
        # TODO: replace ACTION_CANCEL_FLOW (ATO-2097)
        result = self._prediction_result(ACTION_CANCEL_FLOW, domain)
        stack = tracker.stack
        if not stack.is_empty():
            stack.pop()
            stack.push(pattern_stack_frame)
        events: List[Event] = tracker.create_stack_updated_events(stack)
        return self._prediction(result, action_metadata=None, events=events)

    def _prediction_result(
        self, action_name: Optional[Text], domain: Domain, score: Optional[float] = 1.0
    ) -> List[float]:
        """Creates a prediction result.

        Args:
            action_name: The name of the predicted action.
            domain: The model's domain.
            score: The score of the predicted action.

        Returns:
        The prediction result where the score is used for one hot encoding.
        """
        result = self._default_predictions(domain)
        if action_name:
            result[domain.index_for_action(action_name)] = score  # type: ignore[assignment]
        return result

    @classmethod
    def load(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        **kwargs: Any,
    ) -> "EnterpriseSearchPolicy":
        """Loads a trained policy (see parent class for full docstring)."""
        # Perform health checks for both LLM and embeddings client configs
        cls._perform_health_checks(config, "enterprise_search_policy.load")

        prompt_template = None
        try:
            with model_storage.read_from(resource) as path:
                prompt_template = rasa.shared.utils.io.read_file(
                    path / ENTERPRISE_SEARCH_PROMPT_FILE_NAME
                )
        except (FileNotFoundError, FileIOException) as e:
            logger.warning(
                "enterprise_search_policy.load.failed", error=e, resource=resource.name
            )

        store_type = config.get(VECTOR_STORE_PROPERTY, {}).get(
            VECTOR_STORE_TYPE_PROPERTY
        )

        embeddings = cls._create_plain_embedder(config)

        logger.info("enterprise_search_policy.load", config=config)
        if store_type == DEFAULT_VECTOR_STORE_TYPE:
            # if a vector store is not specified,
            # default to using FAISS with the index stored in the model
            # TODO figure out a way to get path without context manager
            with model_storage.read_from(resource) as path:
                vector_store = FAISS_Store(
                    embeddings=embeddings,
                    index_path=path,
                    docs_folder=None,
                    create_index=False,
                )
        else:
            vector_store = create_from_endpoint_config(
                config_type=store_type,
                embeddings=embeddings,
            )  # type: ignore

        return cls(
            config,
            model_storage,
            resource,
            execution_context,
            vector_store=vector_store,
            prompt_template=prompt_template,
        )

    @classmethod
    def _get_local_knowledge_data(cls, config: Dict[str, Any]) -> Optional[List[str]]:
        """This is required only for local knowledge base types.

        e.g. FAISS, to ensure that the graph component is retrained when the knowledge
        base is updated.
        """
        merged_config = {**cls.get_default_config(), **config}

        store_type = merged_config.get(VECTOR_STORE_PROPERTY, {}).get(
            VECTOR_STORE_TYPE_PROPERTY
        )
        if store_type != DEFAULT_VECTOR_STORE_TYPE:
            return None

        source = merged_config.get(VECTOR_STORE_PROPERTY, {}).get(SOURCE_PROPERTY)
        if not source:
            return None

        docs = FAISS_Store.load_documents(source)

        if len(docs) == 0:
            return None

        docs_as_strings = [
            json.dumps(doc.dict(), ensure_ascii=False, sort_keys=True) for doc in docs
        ]
        return sorted(docs_as_strings)

    @classmethod
    def fingerprint_addon(cls, config: Dict[str, Any]) -> Optional[str]:
        """Add a fingerprint of enterprise search policy for the graph."""
        local_knowledge_data = cls._get_local_knowledge_data(config)

        prompt_template = get_prompt_template(
            config.get(PROMPT_CONFIG_KEY),
            DEFAULT_ENTERPRISE_SEARCH_PROMPT_TEMPLATE,
        )

        llm_config = resolve_model_client_config(
            config.get(LLM_CONFIG_KEY), EnterpriseSearchPolicy.__name__
        )
        embedding_config = resolve_model_client_config(
            config.get(EMBEDDINGS_CONFIG_KEY), EnterpriseSearchPolicy.__name__
        )
        return deep_container_fingerprint(
            [prompt_template, local_knowledge_data, llm_config, embedding_config]
        )

    @staticmethod
    def post_process_citations(llm_answer: str) -> str:
        """Post-process the LLM answer.

         Re-writes the bracketed numbers to start from 1 and
         re-arranges the sources to follow the enumeration order.

        Args:
            llm_answer: The LLM answer.

        Returns:
            The post-processed LLM answer.
        """
        logger.debug(
            "enterprise_search_policy.post_process_citations", llm_answer=llm_answer
        )

        # Split llm_answer into answer and citations
        try:
            answer, citations = llm_answer.rsplit("Sources:", 1)
        except ValueError:
            # if there is no "Sources:" in the llm_answer
            return llm_answer

        # Find all source references in the answer
        pattern = r"\[\s*(\d+(?:\s*,\s*\d+)*)\s*\]"
        matches = re.findall(pattern, answer)
        old_source_indices = [
            int(num.strip()) for match in matches for num in match.split(",")
        ]

        # Map old source references to the correct enumeration
        renumber_mapping = {num: idx + 1 for idx, num in enumerate(old_source_indices)}

        # remove whitespace from original source citations in answer
        for match in matches:
            answer = answer.replace(f"[{match}]", f"[{match.replace(' ', '')}]")

        new_answer = []
        for word in answer.split():
            matches = re.findall(pattern, word)
            if matches:
                for match in matches:
                    if "," in match:
                        old_indices = [
                            int(num.strip()) for num in match.split(",") if num
                        ]
                        new_indices = [
                            renumber_mapping[old_index]
                            for old_index in old_indices
                            if old_index in renumber_mapping
                        ]
                        if not new_indices:
                            continue

                        word = word.replace(
                            match, f"{', '.join(map(str, new_indices))}"
                        )
                    else:
                        old_index = int(match.strip("[].,:;?!"))
                        new_index = renumber_mapping.get(old_index)
                        if not new_index:
                            continue

                        word = word.replace(str(old_index), str(new_index))
            new_answer.append(word)

        # join the words
        joined_answer = " ".join(new_answer)
        joined_answer += "\nSources:\n"

        new_sources: List[str] = []

        for line in citations.split("\n"):
            pattern = r"(?<=\[)\d+"
            match = re.search(pattern, line)
            if match:
                old_index = int(match.group(0))
                new_index = renumber_mapping[old_index]
                # replace only the first occurrence of the old index
                line = line.replace(f"[{old_index}]", f"[{new_index}]", 1)

                # insert the line into the new_index position
                new_sources.insert(new_index - 1, line)
            elif line.strip():
                new_sources.append(line)

        joined_sources = "\n".join(new_sources)

        return joined_answer + joined_sources

    @classmethod
    def _perform_health_checks(
        cls, config: Dict[Text, Any], log_source_method: str
    ) -> None:
        # Perform health check of the LLM client config
        llm_config = resolve_model_client_config(config.get(LLM_CONFIG_KEY, {}))
        cls.perform_llm_health_check(
            llm_config,
            DEFAULT_LLM_CONFIG,
            log_source_method,
            EnterpriseSearchPolicy.__name__,
        )

        # Perform health check of the embeddings client config
        embeddings_config = resolve_model_client_config(
            config.get(EMBEDDINGS_CONFIG_KEY, {})
        )
        cls.perform_embeddings_health_check(
            embeddings_config,
            DEFAULT_EMBEDDINGS_CONFIG,
            log_source_method,
            EnterpriseSearchPolicy.__name__,
        )



from typing import Any, Dict, List, Optional, Set

import pandas as pd
import structlog

from rasa.dialogue_understanding.commands import (
    CancelFlowCommand,
    ChitChatAnswerCommand,
    ClarifyCommand,
    HumanHandoffCommand,
    KnowledgeAnswerCommand,
    SetSlotCommand,
    SkipQuestionCommand,
    StartFlowCommand,
)
from rasa.e2e_test.e2e_test_result import TestResult
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.flows.flow_path import FlowPath, FlowPathsList, PathNode

# Report column names
FLOW_NAME_COL_NAME = "Flow Name"
NUM_STEPS_COL_NAME = "Num Steps"
MISSING_STEPS_COL_NAME = "Missing Steps"
LINE_NUMBERS_COL_NAME = "Line Numbers"
COVERAGE_COL_NAME = "Coverage"

FLOWS_KEY = "flows"
NUMBER_OF_STEPS_KEY = "number_of_steps"
NUMBER_OF_UNTESTED_STEPS_KEY = "number_of_untested_steps"
UNTESTED_LINES_KEY = "untested_lines"

SUPPORTED_HISTOGRAM_COMMANDS = [
    KnowledgeAnswerCommand.command(),
    StartFlowCommand.command(),
    SetSlotCommand.command(),
    ClarifyCommand.command(),
    HumanHandoffCommand.command(),
    CancelFlowCommand.command(),
    ChitChatAnswerCommand.command(),
    SkipQuestionCommand.command(),
]

structlogger = structlog.get_logger()


def create_coverage_report(
    flows: FlowsList,
    test_results: List[TestResult],
) -> Optional[pd.DataFrame]:
    """Generates a coverage report.

    This function extracts paths from predefined flows, loads tested paths based
    on the provided e2e test results, and compares the unique nodes of all flow paths
    of one flow to obtain untested nodes. It then generates
    a report that highlights areas of the flows that are not adequately tested.

    Args:
        flows: List of flows.
        test_results: List of e2e test results.

    Returns:
        The coverage report as dataframe.
    """
    if not test_results or flows.user_flows.is_empty():
        return _empty_dataframe()

    # Step 1: Get testable flow paths
    flow_to_testable_paths = flows.extract_flow_paths()

    # Step 2: Get tested flow paths from e2e tests run
    tested_flow_paths = _extract_tested_flow_paths(test_results)
    # No tested flow paths exists, cannot create coverage report
    if not tested_flow_paths:
        return _empty_dataframe()

    # Step 3: Group flow paths by flow
    flow_to_tested_paths = _group_flow_paths_by_flow(tested_flow_paths)

    # Step 4: Get the unvisited nodes and number of unique nodes per flow
    unvisited_nodes_per_flow = _get_unvisited_nodes_per_flow(
        flow_to_testable_paths, flow_to_tested_paths
    )
    number_of_nodes_per_flow = {
        flow: flow_paths.get_number_of_unique_nodes()
        for flow, flow_paths in flow_to_testable_paths.items()
    }

    # Step 5: Produce the report
    coverage_report_data = _create_coverage_report_data(
        flows,
        number_of_nodes_per_flow,
        unvisited_nodes_per_flow,
    )
    return _create_data_frame(coverage_report_data)


def _get_unvisited_nodes_per_flow(
    flow_to_testable_paths: Dict[str, FlowPathsList],
    flow_to_tested_paths: Dict[str, FlowPathsList],
) -> Dict[str, Set[PathNode]]:
    """Returns the unvisited path nodes per flow.

    Compares the set of unique nodes of the testable paths to the unique nodes of the
    tested paths.

    Args:
        flow_to_testable_paths: Testable paths per flow.
        flow_to_tested_paths: Tested paths per flow.

    Returns:
        The unvisited nodes per flow.
    """
    unvisited_nodes_per_flow: Dict[str, Set[PathNode]] = {}

    for flow, testable_paths in flow_to_testable_paths.items():
        if flow in flow_to_tested_paths:
            # get the difference of testable and tested nodes
            testable_nodes = testable_paths.get_unique_nodes()
            tested_nodes = flow_to_tested_paths[flow].get_unique_nodes()
            unvisited_nodes = testable_nodes.difference(tested_nodes)
        else:
            # the flow was not tested at all
            unvisited_nodes = testable_paths.get_unique_nodes()
        unvisited_nodes_per_flow[flow] = unvisited_nodes

    return unvisited_nodes_per_flow


def _group_flow_paths_by_flow(flow_paths: List[FlowPath]) -> Dict[str, FlowPathsList]:
    """Group the available flow paths by flow id.

    Args:
        flow_paths: The list of all flow paths.

    Returns:
        A dictionary mapping a flow to its paths.
    """
    flow_to_paths = {}

    for flow_path in flow_paths:
        if flow_path.flow not in flow_to_paths:
            flow_to_paths[flow_path.flow] = FlowPathsList(flow_path.flow, [flow_path])
        else:
            flow_to_paths[flow_path.flow].paths.append(flow_path)

    return flow_to_paths


def _create_data_frame(coverage_report_data: Dict[str, Any]) -> pd.DataFrame:
    df = _construct_dataframe(coverage_report_data)
    df = _calculate_coverage(df)
    df = _append_total_row(df)
    df = _reorder_columns(df)
    return df


def _empty_dataframe() -> pd.DataFrame:
    """Generates an empty DataFrame.

    The DataFrame includes all required columns for a coverage report.

    Returns:
        pd.DataFrame: An empty DataFrame with predefined columns.
    """
    return pd.DataFrame(
        columns=[
            FLOW_NAME_COL_NAME,
            COVERAGE_COL_NAME,
            NUM_STEPS_COL_NAME,
            MISSING_STEPS_COL_NAME,
            LINE_NUMBERS_COL_NAME,
        ]
    )


def _create_coverage_report_data(
    flows: FlowsList,
    number_of_nodes_per_flow: Dict[str, int],
    unvisited_nodes_per_flow: Dict[str, Set[PathNode]],
) -> Dict[str, Any]:
    """Creates the data for the coverage report.

    Args:
        flows: All available flow names
        number_of_nodes_per_flow: Number of nodes per flow
        unvisited_nodes_per_flow: Unvisited nodes per flow

    Returns:
        A dictionary with processed data needed to construct the DataFrame.
    """
    flow_ids = [flow.id for flow in flows.user_flows.underlying_flows]

    flow_full_names = ["unknown"] * len(flow_ids)
    number_of_steps = [0] * len(flow_ids)
    number_of_untested_steps = [0] * len(flow_ids)
    untested_lines: List[List[str]] = [[]] * len(flow_ids)

    for flow in flow_ids:
        nodes = unvisited_nodes_per_flow[flow]
        lines: List[str] = [node.lines for node in nodes if node.lines]

        index = flow_ids.index(flow)
        if flow_object := flows.flow_by_id(flow):
            flow_full_names[index] = flow_object.get_full_name()
        number_of_steps[index] = number_of_nodes_per_flow[flow]
        number_of_untested_steps[index] = len(nodes)
        untested_lines[index] = lines

    return {
        FLOWS_KEY: flow_full_names,
        NUMBER_OF_STEPS_KEY: number_of_steps,
        NUMBER_OF_UNTESTED_STEPS_KEY: number_of_untested_steps,
        UNTESTED_LINES_KEY: _reformat_untested_lines(untested_lines),
    }


def _reformat_untested_lines(untested_lines: List[List[str]]) -> List[str]:
    """Format lists of lists to list of str.

    Formats nested lists of untested lines into a string format suitable
    for display.

    Args:
        untested_lines: A list of lists containing line information.

    Returns:
        A list of formatted strings representing untested lines.
    """
    return ["[" + ", ".join(sublist) + "]" for sublist in untested_lines]


def _construct_dataframe(report_data: Dict[str, Any]) -> pd.DataFrame:
    """Constructs a DataFrame from the provided data.

    Args:
        report_data: A dictionary containing report data.

    Returns:
        pd.DataFrame: A DataFrame constructed from the report data.
    """
    return pd.DataFrame(
        {
            FLOW_NAME_COL_NAME: report_data[FLOWS_KEY],
            NUM_STEPS_COL_NAME: report_data[NUMBER_OF_STEPS_KEY],
            MISSING_STEPS_COL_NAME: report_data[NUMBER_OF_UNTESTED_STEPS_KEY],
            LINE_NUMBERS_COL_NAME: report_data[UNTESTED_LINES_KEY],
        }
    )


def _calculate_coverage(df: pd.DataFrame) -> pd.DataFrame:
    """Calculates the coverage percentage and updates the DataFrame.

    Args:
        df: The DataFrame to update with coverage data.

    Returns:
        The updated DataFrame with coverage percentages.
    """
    df[COVERAGE_COL_NAME] = (
        (df[NUM_STEPS_COL_NAME] - df[MISSING_STEPS_COL_NAME]) / df[NUM_STEPS_COL_NAME]
    ) * 100

    # Set the float format for displaying coverage percentages
    pd.options.display.float_format = "{:.2f}%".format

    return df


def _append_total_row(df: pd.DataFrame) -> pd.DataFrame:
    """Appends a total summary row to the DataFrame.

    Args:
        df: The DataFrame to which the total row will be appended.

    Returns:
        The updated DataFrame with an appended total row.
    """
    total_data = {
        FLOW_NAME_COL_NAME: "Total",
        NUM_STEPS_COL_NAME: df[NUM_STEPS_COL_NAME].sum(),
        MISSING_STEPS_COL_NAME: df[MISSING_STEPS_COL_NAME].sum(),
        LINE_NUMBERS_COL_NAME: "",
        COVERAGE_COL_NAME: _calculate_total_coverage(df),
    }
    # Append the total row using `.loc`
    df.loc[len(df)] = total_data

    return df


def _calculate_total_coverage(df: pd.DataFrame) -> str:
    """Calculates the total coverage percentage for the DataFrame.

    Args:
        df: The DataFrame for which total coverage is calculated.

    Returns:
        The calculated total coverage percentage formatted as a string.
    """
    total_steps = df[NUM_STEPS_COL_NAME].sum()
    missing_steps = df[MISSING_STEPS_COL_NAME].sum()
    return (
        pd.NA
        if total_steps == 0
        else f"{((total_steps - missing_steps) / total_steps) * 100:.2f}%"
    )


def _reorder_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Reorders the columns of the DataFrame to prioritize specific columns.

    Args:
        df: The DataFrame whose columns are to be reordered.

    Returns:
        The DataFrame with reordered columns.
    """
    columns = [FLOW_NAME_COL_NAME, COVERAGE_COL_NAME] + [
        col for col in df.columns if col not in [FLOW_NAME_COL_NAME, COVERAGE_COL_NAME]
    ]
    return df[columns]


def _extract_tested_flow_paths(test_results: List[TestResult]) -> List[FlowPath]:
    """Extract the flow paths of the test results.

    Args:
        test_results: List of test results.

    Returns:
        List[FlowPath]: A list of FlowPaths.
    """
    flatten_paths = []

    for test_result in test_results:
        if test_result.tested_paths:
            for path in test_result.tested_paths:
                flatten_paths.append(path)

    return flatten_paths


def extract_tested_commands(test_results: List[TestResult]) -> Dict[str, int]:
    """Extract tested commands from the test results.

    Args:
        test_results: List of test results.

    Returns:
        Dict[str, int]: A dictionary of commands and their counts.
    """
    command_histogram_data = {}
    for command in SUPPORTED_HISTOGRAM_COMMANDS:
        command_histogram_data[command] = 0

    for test_result in test_results:
        if test_result.tested_commands:
            for flow, commands_dict in test_result.tested_commands.items():
                for command, count in commands_dict.items():
                    if command in command_histogram_data:
                        command_histogram_data[command] += count

    return dict(command_histogram_data)



import json
import uuid
from dataclasses import asdict, dataclass, field
from typing import Any, Awaitable, Callable, Dict, List, Text

import structlog
from sanic import Websocket  # type: ignore[attr-defined]

from rasa.core.channels.channel import UserMessage
from rasa.core.channels.voice_ready.utils import CallParameters

structlogger = structlog.get_logger()


@dataclass
class NewSessionMessage:
    """Message indicating a new session has been started."""

    call_sid: str
    message_id: str
    call_params: CallParameters

    @staticmethod
    def from_message(message: Dict[str, Any]) -> "NewSessionMessage":
        structlogger.debug("jambonz.websocket.message.new_session", message=message)
        call_params = CallParameters(
            call_id=message.get("call_sid"),
            user_phone=message.get("data", {}).get("from"),
            bot_phone=message.get("data", {}).get("to"),
        )
        return NewSessionMessage(
            message.get("call_sid"),
            message.get("msgid"),
            call_params,
        )


@dataclass
class Transcript:
    """Transcript of a spoken utterance."""

    text: str
    confidence: float


@dataclass
class TranscriptResult:
    """Result of an ASR call with potential transcripts."""

    call_sid: str
    message_id: str
    is_final: bool
    transcripts: List[Transcript] = field(default_factory=list)

    @staticmethod
    def from_speech_result(message: Dict[str, Any]) -> "TranscriptResult":
        return TranscriptResult(
            message.get("call_sid"),
            message.get("msgid"),
            message.get("data", {}).get("speech", {}).get("is_final", True),
            transcripts=[
                Transcript(t.get("transcript", ""), t.get("confidence", 1.0))
                for t in message.get("data", {})
                .get("speech", {})
                .get("alternatives", [])
            ],
        )

    @staticmethod
    def from_dtmf_result(message: Dict[str, Any]) -> "TranscriptResult":
        """Create a transcript result from a DTMF result.

        We use the dtmf as the text with confidence 1.0
        """
        return TranscriptResult(
            message.get("call_sid"),
            message.get("msgid"),
            is_final=True,
            transcripts=[
                Transcript(str(message.get("data", {}).get("digits", "")), 1.0)
            ],
        )


@dataclass
class CallStatusChanged:
    """Message indicating a change in the call status."""

    call_sid: str
    status: str

    @staticmethod
    def from_message(message: Dict[str, Any]) -> "CallStatusChanged":
        structlogger.debug(
            "jambonz.websocket.message.call_status_changed",
            message=message,
        )
        return CallStatusChanged(
            message.get("call_sid"), message.get("data", {}).get("call_status")
        )


@dataclass
class SessionReconnect:
    """Message indicating a session has reconnected."""

    call_sid: str

    @staticmethod
    def from_message(message: Dict[str, Any]) -> "SessionReconnect":
        return SessionReconnect(message.get("call_sid"))


@dataclass
class VerbStatusChanged:
    """Message indicating a change in the status of a verb."""

    call_sid: str
    event: str
    id: str
    name: str

    @staticmethod
    def from_message(message: Dict[str, Any]) -> "VerbStatusChanged":
        return VerbStatusChanged(
            message.get("call_sid"),
            message.get("data", {}).get("event"),
            message.get("data", {}).get("id"),
            message.get("data", {}).get("name"),
        )


@dataclass
class GatherTimeout:
    """Message indicating a gather timeout."""

    call_sid: str

    @staticmethod
    def from_message(message: Dict[str, Any]) -> "GatherTimeout":
        return GatherTimeout(message.get("call_sid"))


async def websocket_message_handler(
    message_dump: str,
    on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ws: Websocket,
) -> None:
    """Handle incoming messages from the websocket."""
    message = json.loads(message_dump)

    # parse and handle the different message types
    if message.get("type") == "session:new":
        new_session = NewSessionMessage.from_message(message)
        await handle_new_session(new_session, on_new_message, ws)
    elif message.get("type") == "session:reconnect":
        session_reconnect = SessionReconnect.from_message(message)
        await handle_session_reconnect(session_reconnect)
    elif message.get("type") == "call:status":
        call_status = CallStatusChanged.from_message(message)
        await handle_call_status(call_status, on_new_message, ws)
    elif message.get("type") == "verb:hook" and message.get("hook") == "/gather":
        hook_trigger_reason = message.get("data", {}).get("reason")

        if hook_trigger_reason == "speechDetected":
            transcript = TranscriptResult.from_speech_result(message)
            await handle_gather_completed(transcript, on_new_message, ws)
        elif hook_trigger_reason == "timeout":
            gather_timeout = GatherTimeout.from_message(message)
            await handle_gather_timeout(gather_timeout, ws)
        elif hook_trigger_reason == "dtmfDetected":
            # for now, let's handle it as normal user input with a
            # confidence of 1.0
            transcript = TranscriptResult.from_dtmf_result(message)
            await handle_gather_completed(transcript, on_new_message, ws)
        else:
            structlogger.debug(
                "jambonz.websocket.message.verb_hook",
                call_sid=message.get("call_sid"),
                reason=hook_trigger_reason,
                message=message,
            )
    elif message.get("type") == "verb:status":
        verb_status = VerbStatusChanged.from_message(message)
        await handle_verb_status(verb_status)
    elif message.get("type") == "jambonz:error":
        # jambonz ran into a fatal error handling the call. the call will be
        # terminated.
        structlogger.error("jambonz.websocket.message.error", message=message)
    else:
        structlogger.warning("jambonz.websocket.message.unknown_type", message=message)


async def handle_new_session(
    message: NewSessionMessage,
    on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ws: Websocket,
) -> None:
    """Handle new session message."""
    from rasa.core.channels.voice_ready.jambonz import JambonzWebsocketOutput

    structlogger.debug("jambonz.websocket.message.new_call", call_sid=message.call_sid)
    output_channel = JambonzWebsocketOutput(ws, message.call_sid)
    user_msg = UserMessage(
        text="/session_start",
        output_channel=output_channel,
        sender_id=message.call_sid,
        metadata=asdict(message.call_params),
    )
    await send_config_ack(message.message_id, ws)
    await on_new_message(user_msg)
    await send_gather_input(ws)


async def handle_gather_completed(
    transcript_result: TranscriptResult,
    on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ws: Websocket,
) -> None:
    """Handle changes to commands we have send to jambonz.

    This includes results of gather calles with their transcription.
    """
    from rasa.core.channels.voice_ready.jambonz import JambonzWebsocketOutput

    if not transcript_result.is_final:
        # in case of a non final transcript, we are going to wait for the final
        # one and ignore the partial one
        structlogger.debug(
            "jambonz.websocket.message.transcript_partial",
            call_sid=transcript_result.call_sid,
            number_of_transcripts=len(transcript_result.transcripts),
        )
        return

    if transcript_result.transcripts:
        most_likely_transcript = transcript_result.transcripts[0]
        output_channel = JambonzWebsocketOutput(ws, transcript_result.call_sid)
        user_msg = UserMessage(
            text=most_likely_transcript.text,
            output_channel=output_channel,
            sender_id=transcript_result.call_sid,
            metadata={},
        )
        structlogger.debug(
            "jambonz.websocket.message.transcript",
            call_sid=transcript_result.call_sid,
            transcript=most_likely_transcript.text,
            confidence=most_likely_transcript.confidence,
            number_of_transcripts=len(transcript_result.transcripts),
        )
        await on_new_message(user_msg)
    else:
        structlogger.warning(
            "jambonz.websocket.message.no_transcript",
            call_sid=transcript_result.call_sid,
        )
    await send_gather_input(ws)


async def handle_gather_timeout(gather_timeout: GatherTimeout, ws: Websocket) -> None:
    """Handle gather timeout."""
    structlogger.debug(
        "jambonz.websocket.message.gather_timeout",
        call_sid=gather_timeout.call_sid,
    )
    # TODO: figure out how to handle timeouts
    await send_ws_text_message(ws, "I'm sorry, I didn't catch that.")
    await send_gather_input(ws)


async def handle_call_status(
    call_status: CallStatusChanged,
    on_new_message: Callable[[UserMessage], Awaitable[Any]],
    ws: Websocket,
) -> None:
    """Handle changes in the call status."""
    structlogger.debug(
        "jambonz.websocket.message.call_status_changed",
        call_sid=call_status.call_sid,
        message=call_status.status,
    )

    if call_status.status == "completed":
        structlogger.debug("jambonz.websocket.message.call_completed")
        from rasa.core.channels.voice_ready.jambonz import JambonzWebsocketOutput

        output_channel = JambonzWebsocketOutput(ws, call_status.call_sid)
        user_msg = UserMessage(
            text="/session_end",
            output_channel=output_channel,
            sender_id=call_status.call_sid,
            metadata={},
        )
        await on_new_message(user_msg)


async def handle_session_reconnect(session_reconnect: SessionReconnect) -> None:
    """Handle session reconnect message."""
    # there is nothing we need to do atm when a session reconnects.
    # this happens if jambonz looses the websocket connection and reconnects
    structlogger.debug(
        "jambonz.websocket.message.session_reconnect",
        call_sid=session_reconnect.call_sid,
    )


async def handle_verb_status(verb_status: VerbStatusChanged) -> None:
    """Handle changes in the status of a verb."""
    structlogger.debug(
        "jambonz.websocket.message.verb_status_changed",
        call_sid=verb_status.call_sid,
        event_type=verb_status.event,
        id=verb_status.id,
        name=verb_status.name,
    )


async def send_config_ack(message_id: str, ws: Websocket) -> None:
    """Send an ack message to jambonz including the configuration."""
    await ws.send(
        json.dumps(
            {
                "type": "ack",
                "msgid": message_id,
                "data": [{"config": {"notifyEvents": True}}],
            }
        )
    )


async def send_gather_input(ws: Websocket) -> None:
    """Send a gather input command to jambonz."""
    structlogger.debug("jambonz.websocket.send.gather")
    await ws.send(
        json.dumps(
            {
                "type": "command",
                "command": "redirect",
                "queueCommand": True,
                "data": [
                    {
                        "gather": {
                            "input": ["speech", "digits"],
                            "minDigits": 1,
                            "id": uuid.uuid4().hex,
                            "actionHook": "/gather",
                        }
                    }
                ],
            }
        )
    )


async def send_ws_text_message(ws: Websocket, text: Text) -> None:
    """Send a text message to the websocket using the jambonz interface."""
    await ws.send(
        json.dumps(
            {
                "type": "command",
                "command": "redirect",
                "queueCommand": True,
                "data": [
                    {
                        "say": {
                            # id can be used for status notifications
                            "id": uuid.uuid4().hex,
                            "text": text,
                        }
                    }
                ],
            }
        )
    )


async def send_ws_hangup_message(hangup_delay_seconds: int, ws: Websocket) -> None:
    """Send a hangup message to the websocket using the jambonz interface."""
    structlogger.debug("jambonz.websocket.send.hangup")
    await ws.send(
        json.dumps(
            {
                "type": "command",
                "command": "redirect",
                "queueCommand": True,
                "data": [
                    {"pause": {"length": hangup_delay_seconds}},
                    {
                        "hangup": {},
                    },
                ],
            }
        )
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional, Set, Text

from rasa.shared.constants import UTTER_PREFIX
from rasa.shared.core.flows.flow_step import FlowStep


@dataclass
class ActionFlowStep(FlowStep):
    """A flow step that that defines an action to be executed."""

    action: Text
    """The action of the flow step."""

    @classmethod
    def from_json(cls, flow_id: Text, data: Dict[Text, Any]) -> ActionFlowStep:
        """Create an ActionFlowStep object from serialized data

        Args:
            flow_id: The id of the flow that contains the step.
            data: data for an ActionFlowStep object in a serialized format

        Returns:
            An ActionFlowStep object
        """
        base = super().from_json(flow_id, data)
        return ActionFlowStep(
            action=data["action"],
            **base.__dict__,
        )

    def as_json(self) -> Dict[Text, Any]:
        """Serialize the ActionFlowStep

        Returns:
            The ActionFlowStep object as serialized data.
        """
        data = super().as_json()
        data["action"] = self.action
        return data

    @property
    def default_id_postfix(self) -> str:
        """Returns the default id postfix of the flow step."""
        return self.action

    @property
    def utterances(self) -> Set[str]:
        """Return all the utterances used in this step."""
        return {self.action} if self.action.startswith(UTTER_PREFIX) else set()

    @property
    def custom_action(self) -> Optional[str]:
        """Return all the custom actions used in this step."""
        return self.action if not self.action.startswith(UTTER_PREFIX) else None



import datetime
import logging
import uuid
from typing import Any, AsyncIterator, Dict, List, Optional, Set, Text

from tqdm import tqdm

import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.core.brokers.broker import EventBroker
from rasa.core.brokers.pika import PikaEventBroker
from rasa.core.constants import RASA_EXPORT_PROCESS_ID_HEADER_NAME
from rasa.core.tracker_store import TrackerStore
from rasa.exceptions import (
    NoConversationsInTrackerStoreError,
    NoEventsToMigrateError,
    PublishingError,
)
from rasa.shared.core.trackers import EventVerbosity

logger = logging.getLogger(__name__)


class Exporter:
    """Manages the publishing of events in a tracker store to an event broker.

    Attributes:
        endpoints_path: Path to the endpoints file used to configure the event
            broker and tracker store. If `None`, the default path ('endpoints.yml')
            is used.
        tracker_store: `TrackerStore` to export conversations from.
        event_broker: `EventBroker` to export conversations to.
        requested_conversation_ids: List of conversation IDs requested to be
            processed.
        minimum_timestamp: Minimum timestamp of events that are published.
            If `None`, apply no such constraint.
        maximum_timestamp: Maximum timestamp of events that are published.
            If `None`, apply no such constraint.
    """

    def __init__(
        self,
        tracker_store: TrackerStore,
        event_broker: EventBroker,
        endpoints_path: Text,
        requested_conversation_ids: Optional[Text] = None,
        minimum_timestamp: Optional[float] = None,
        maximum_timestamp: Optional[float] = None,
        offset_timestamps_by_seconds: Optional[int] = None,
    ) -> None:
        self.endpoints_path = endpoints_path
        self.tracker_store = tracker_store

        self.event_broker = event_broker
        self.requested_conversation_ids = requested_conversation_ids
        self.minimum_timestamp = minimum_timestamp
        self.maximum_timestamp = maximum_timestamp
        self.offset_timestamps_by_seconds = offset_timestamps_by_seconds

    async def publish_events(self) -> int:
        """Publish events in a tracker store using an event broker.

        Exits if the publishing of events is interrupted due to an error. In that case,
        the CLI command to continue the export where it was interrupted is printed.

        Returns:
            The number of successfully published events.
        """
        self._print_offset_info()

        published_events = 0
        current_timestamp = None

        headers = self._get_message_headers()

        async for event in self._fetch_events_within_time_range():
            # noinspection PyBroadException
            try:
                self._publish_with_message_headers(event, headers)
                published_events += 1
                current_timestamp = event["timestamp"]
            except Exception as e:
                logger.exception(e)
                raise PublishingError(current_timestamp)

        await self.event_broker.close()

        return published_events

    def _print_offset_info(self) -> None:
        """Output information about the offset applied to event timestamps."""
        if self.offset_timestamps_by_seconds is None:
            return

        delta = datetime.timedelta(seconds=abs(self.offset_timestamps_by_seconds))
        operator = "-" if self.offset_timestamps_by_seconds > 0 else ""
        rasa.shared.utils.cli.print_info(
            f"All event timestamps will be offset by {operator}{delta}! ⏰"
        )

    def _get_message_headers(self) -> Optional[Dict[Text, Text]]:
        """Generate a message header for publishing events to a `PikaEventBroker`.

        Returns:
            Message headers with a randomly generated uuid under the
            `RASA_EXPORT_PROCESS_ID_HEADER_NAME` key if `self.event_broker` is a
            `PikaEventBroker`, else `None`.

        """
        if isinstance(self.event_broker, PikaEventBroker):
            return {RASA_EXPORT_PROCESS_ID_HEADER_NAME: uuid.uuid4().hex}

        return None

    def _publish_with_message_headers(
        self, original_event: Dict[Text, Any], headers: Optional[Dict[Text, Text]]
    ) -> None:
        """Publish `event` to a message broker with `headers`.

        Args:
            original_event: Serialized event to be published.
            headers: Message headers to be published if `self.event_broker` is a
                `PikaEventBroker`.

        """
        if self.offset_timestamps_by_seconds is not None:
            event = dict(original_event)
            event["timestamp"] += self.offset_timestamps_by_seconds
        else:
            event = original_event

        if isinstance(self.event_broker, PikaEventBroker):
            self.event_broker.publish(event=event, headers=headers)
        else:
            self.event_broker.publish(event)

    async def _get_conversation_ids_in_tracker(self) -> Set[Text]:
        """Fetch conversation IDs in `self.tracker_store`.

        Returns:
            A set of conversation IDs in `self.tracker_store`.

        Raises:
            `NoConversationsInTrackerStoreError` if
            `conversation_ids_in_tracker_store` is empty.

        """
        conversation_ids_in_tracker_store = set(await self.tracker_store.keys())

        if conversation_ids_in_tracker_store:
            return conversation_ids_in_tracker_store

        raise NoConversationsInTrackerStoreError(
            "Could not find any conversations in connected tracker store. "
            "Please validate your `endpoints.yml` and make sure the defined "
            "tracker store exists. Exiting."
        )

    def _validate_all_requested_ids_exist(
        self, conversation_ids_in_tracker_store: Set[Text]
    ) -> None:
        """Warn user if `self.requested_conversation_ids` contains IDs not found in
        `conversation_ids_in_tracker_store`.

        Args:
            conversation_ids_in_tracker_store: Set of conversation IDs contained in
            the tracker store.

        """
        missing_ids_in_tracker_store = (
            set(self.requested_conversation_ids) - conversation_ids_in_tracker_store
        )
        if missing_ids_in_tracker_store:
            rasa.shared.utils.cli.print_warning(
                f"Could not find the following requested "
                f"conversation IDs in connected tracker store: "
                f"{', '.join(sorted(missing_ids_in_tracker_store))}"
            )

    async def _get_conversation_ids_to_process(self) -> Set[Text]:
        """Get conversation IDs that are good for processing.

        Finds the intersection of events that are contained in the tracker store with
        those events requested as a command-line argument.

        Returns:
            Conversation IDs that are both requested and contained in the tracker
            store. If no conversation IDs are requested, all conversation IDs in the
            tracker store are returned.

        """
        conversation_ids_in_tracker_store = (
            await self._get_conversation_ids_in_tracker()
        )

        if not self.requested_conversation_ids:
            return conversation_ids_in_tracker_store

        self._validate_all_requested_ids_exist(conversation_ids_in_tracker_store)

        conversation_ids_to_process = conversation_ids_in_tracker_store & set(
            self.requested_conversation_ids
        )

        if not conversation_ids_to_process:
            raise NoEventsToMigrateError(
                "Could not find an overlap between the requested "
                "conversation IDs and those found in the tracker store. Exiting."
            )

        return conversation_ids_to_process

    async def _fetch_events_within_time_range(self) -> AsyncIterator[Dict[Text, Any]]:
        """Fetch all events for `conversation_ids` within the supplied time range.

        Returns:
            Serialized events with added `sender_id` field.

        """
        conversation_ids_to_process = await self._get_conversation_ids_to_process()

        rasa.shared.utils.cli.print_info(
            f"Fetching events for {len(conversation_ids_to_process)} "
            f"conversation IDs:"
        )
        for conversation_id in tqdm(conversation_ids_to_process, "conversation IDs"):
            tracker = await self.tracker_store.retrieve_full_tracker(conversation_id)
            if not tracker:
                logger.info(
                    f"Could not retrieve tracker for conversation ID "
                    f"'{conversation_id}'. Skipping."
                )
                continue

            _events = tracker.current_state(EventVerbosity.ALL)["events"]

            if not _events:
                logger.info(
                    f"No events to migrate for conversation ID '{conversation_id}'."
                )
                continue

            events = self._get_events_for_conversation_id(_events, conversation_id)

            # the order of events was changed after ATO-2192
            # more context: https://github.com/RasaHQ/rasa/pull/13019
            # we should sort the events by timestamp to keep the order
            events.sort(key=lambda x: x["timestamp"])

            # the conversation IDs are needed in the event publishing
            for event in events:
                if (
                    self.minimum_timestamp is not None
                    and event["timestamp"] < self.minimum_timestamp
                ):
                    continue
                if (
                    self.maximum_timestamp is not None
                    and event["timestamp"] >= self.maximum_timestamp
                ):
                    continue
                yield event

    @staticmethod
    def _get_events_for_conversation_id(
        events: List[Dict[Text, Any]], conversation_id: Text
    ) -> List[Dict[Text, Any]]:
        """Get serialised events with added `sender_id` key.

        Args:
            events: Events to modify.
            conversation_id: Conversation ID to add to events.

        Returns:
            Events with added `sender_id` key.

        """
        events_with_conversation_id = []

        for event in events:
            event["sender_id"] = conversation_id
            events_with_conversation_id.append(event)

        return events_with_conversation_id



import asyncio
import filecmp
import logging
import os
import re
import tempfile
import warnings
from asyncio import AbstractEventLoop
from collections import OrderedDict
from io import StringIO
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Text,
    Type,
    Union,
)

from ruamel import yaml
from ruamel.yaml import RoundTripRepresenter
from typing_extensions import Protocol

import rasa.shared.constants
import rasa.shared.utils.io

if TYPE_CHECKING:
    from prompt_toolkit.validation import Validator

YAML_LINE_MAX_WIDTH = 4096


class WriteRow(Protocol):
    """Describes a csv writer supporting a `writerow` method (workaround for typing)."""

    def writerow(self, row: List[Text]) -> None:
        """Write the given row.

        Args:
            row: the entries of a row as a list of strings
        """
        ...


def configure_colored_logging(loglevel: Text) -> None:
    """Configures coloredlogs library for specified loglevel.

    Args:
        loglevel: The loglevel to configure the library for
    """
    import coloredlogs

    loglevel = loglevel or os.environ.get(
        rasa.shared.constants.ENV_LOG_LEVEL, rasa.shared.constants.DEFAULT_LOG_LEVEL
    )

    field_styles = coloredlogs.DEFAULT_FIELD_STYLES.copy()
    field_styles["asctime"] = {}
    level_styles = coloredlogs.DEFAULT_LEVEL_STYLES.copy()
    level_styles["debug"] = {}
    coloredlogs.install(
        level=loglevel,
        use_chroot=False,
        fmt="%(asctime)s %(levelname)-8s %(name)s  - %(message)s",
        level_styles=level_styles,
        field_styles=field_styles,
    )


def enable_async_loop_debugging(
    event_loop: AbstractEventLoop, slow_callback_duration: float = 0.1
) -> AbstractEventLoop:
    """Enables debugging on an event loop.

    Args:
        event_loop: The event loop to enable debugging on
        slow_callback_duration: The threshold at which a callback should be
                                alerted as slow.
    """
    logging.info(
        "Enabling coroutine debugging. Loop id {}.".format(id(asyncio.get_event_loop()))
    )

    # Enable debugging
    event_loop.set_debug(True)

    # Make the threshold for "slow" tasks very very small for
    # illustration. The default is 0.1 (= 100 milliseconds).
    event_loop.slow_callback_duration = slow_callback_duration

    # Report all mistakes managing asynchronous resources.
    warnings.simplefilter("always", ResourceWarning)
    return event_loop


def create_temporary_file(data: Any, suffix: Text = "", mode: Text = "w+") -> Text:
    """Creates a tempfile.NamedTemporaryFile object for data."""
    encoding = None if "b" in mode else rasa.shared.utils.io.DEFAULT_ENCODING
    f = tempfile.NamedTemporaryFile(
        mode=mode, suffix=suffix, delete=False, encoding=encoding
    )
    f.write(data)

    f.close()
    return f.name


def create_temporary_directory() -> Text:
    """Creates a tempfile.TemporaryDirectory."""
    f = tempfile.TemporaryDirectory()
    return f.name


def create_path(file_path: Text) -> None:
    """Makes sure all directories in the 'file_path' exists."""
    parent_dir = os.path.dirname(os.path.abspath(file_path))
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir)


def file_type_validator(
    valid_file_types: List[Text], error_message: Text
) -> Type["Validator"]:
    """Creates a `Validator` class which can be used with `questionary` to validate
    file paths.
    """

    def is_valid(path: Text) -> bool:
        return path is not None and any(
            [path.endswith(file_type) for file_type in valid_file_types]
        )

    return create_validator(is_valid, error_message)


def not_empty_validator(error_message: Text) -> Type["Validator"]:
    """Creates a `Validator` class which can be used with `questionary` to validate
    that the user entered something other than whitespace.
    """

    def is_valid(input: Text) -> bool:
        return input is not None and input.strip() != ""

    return create_validator(is_valid, error_message)


def create_validator(
    function: Callable[[Text], bool], error_message: Text
) -> Type["Validator"]:
    """Helper method to create `Validator` classes from callable functions. Should be
    removed when questionary supports `Validator` objects.
    """
    from prompt_toolkit.document import Document
    from prompt_toolkit.validation import ValidationError, Validator

    class FunctionValidator(Validator):
        @staticmethod
        def validate(document: Document) -> None:
            is_valid = function(document.text)
            if not is_valid:
                raise ValidationError(message=error_message)

    return FunctionValidator


def remove_emojis(s: str) -> str:
    import demoji

    replaced = demoji.replace(s)
    if replaced == s:
        return s
    # remove duplicate or trailing whitespaces if emojis were removed
    return re.sub(r" +", " ", replaced).strip()


def are_directories_equal(dir1: Path, dir2: Path) -> bool:
    """Compares two directories recursively.

    Files in each directory are
    assumed to be equal if their names and contents are equal.

    Args:
        dir1: The first directory.
        dir2: The second directory.

    Returns:
        `True` if they are equal, `False` otherwise.
    """
    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if dirs_cmp.left_only or dirs_cmp.right_only:
        return False

    (_, mismatches, errors) = filecmp.cmpfiles(
        dir1, dir2, dirs_cmp.common_files, shallow=False
    )

    if mismatches or errors:
        return False

    for common_dir in dirs_cmp.common_dirs:
        new_dir1 = Path(dir1, common_dir)
        new_dir2 = Path(dir2, common_dir)

        is_equal = are_directories_equal(new_dir1, new_dir2)
        if not is_equal:
            return False

    return True


def write_yaml(
    data: Any,
    target: Union[Text, Path, StringIO],
    transform: Optional[Callable] = None,
    should_preserve_key_order: bool = False,
) -> None:
    """Writes a yaml to the file or to the stream.

    Args:
        data: The data to write.
        target: The path to the file which should be written or a stream object
        transform: A function to transform the data before writing it.
        should_preserve_key_order: Whether to force preserve key order in `data`.
    """
    yaml.add_representer(
        OrderedDict,
        RoundTripRepresenter.represent_dict,
        representer=RoundTripRepresenter,
    )

    if should_preserve_key_order:
        data = rasa.shared.utils.io.convert_to_ordered_dict(data)

    dumper = yaml.YAML()
    # no wrap lines
    dumper.width = YAML_LINE_MAX_WIDTH  # type: ignore

    # use `null` to represent `None`
    dumper.representer.add_representer(
        type(None),
        lambda self, _: self.represent_scalar("tag:yaml.org,2002:null", "null"),
    )

    if isinstance(target, StringIO):
        dumper.dump(data, target, transform=transform)
        return

    with Path(target).open("w", encoding="utf-8") as outfile:
        dumper.dump(data, outfile, transform=transform)



from typing import Optional

import structlog
from jinja2 import Template

from rasa.core.tracker_store import DialogueStateTracker
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.utils.llm import (
    tracker_as_readable_transcript,
)

structlogger = structlog.get_logger()

_DEFAULT_SUMMARIZER_TEMPLATE = """Summarize the provided conversation between
a user and a conversational AI. The summary should be a short text that
captures the main points of the conversation.

Conversation:
{{conversation}}

Summary:"""
SUMMARY_PROMPT_TEMPLATE = Template(_DEFAULT_SUMMARIZER_TEMPLATE)
MAX_TURNS_DEFAULT = 20


def _create_summarization_prompt(
    tracker: DialogueStateTracker, max_turns: Optional[int]
) -> str:
    """Creates an LLM prompt to summarize the conversation in the tracker.

    Args:
        tracker: tracker of the conversation to be summarized
        max_turns: maximum number of turns to summarize


    Returns:
        The prompt to summarize the conversation.
    """
    transcript = tracker_as_readable_transcript(tracker, max_turns=max_turns)
    return SUMMARY_PROMPT_TEMPLATE.render(
        conversation=transcript,
    )


async def summarize_conversation(
    tracker: DialogueStateTracker,
    llm: LLMClient,
    max_turns: Optional[int] = MAX_TURNS_DEFAULT,
) -> str:
    """Summarizes the dialogue using the LLM.

    Args:
        tracker: the tracker to summarize
        llm: the LLM to use for summarization
        max_turns: maximum number of turns to summarize

    Returns:
        The summary of the dialogue.
    """
    prompt = _create_summarization_prompt(tracker, max_turns)
    try:
        llm_response = await llm.acompletion(prompt)
        summarization = llm_response.choices[0].strip()
        structlogger.debug(
            "summarization.success", summarization=summarization, prompt=prompt
        )
        return summarization
    except Exception as e:
        transcript = tracker_as_readable_transcript(tracker, max_turns=max_turns)
        structlogger.error("summarization.error", error=e)
        return transcript



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, Dict, List

import structlog

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxManager,
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.patterns.human_handoff import (
    HumanHandoffPatternFlowStackFrame,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class HumanHandoffCommand(Command):
    """A command to indicate that the bot should handoff to a human."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "human handoff"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> HumanHandoffCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return HumanHandoffCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        stack.push(HumanHandoffPatternFlowStackFrame())
        structlogger.debug(
            "command_executor.human_handoff.pushed_to_stack", command=self
        )
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, HumanHandoffCommand)

    def to_dsl(self) -> str:
        """Converts the command to a DSL string."""
        mapper = {
            CommandSyntaxVersion.v1: "HumanHandoff()",
            CommandSyntaxVersion.v2: "hand over",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )

    @classmethod
    def from_dsl(cls, match: re.Match, **kwargs: Any) -> HumanHandoffCommand:
        """Converts the DSL string to a command."""
        return HumanHandoffCommand()

    @staticmethod
    def regex_pattern() -> str:
        mapper = {
            CommandSyntaxVersion.v1: r"HumanHandoff\(\)",
            CommandSyntaxVersion.v2: r"^[^\w]*hand over$",
        }
        return mapper.get(
            CommandSyntaxManager.get_syntax_version(),
            mapper[CommandSyntaxManager.get_default_syntax_version()],
        )



import importlib.resources
import re
from typing import Any, Dict, List, Optional, Tuple

import structlog
from jinja2 import Template

from rasa.llm_fine_tuning.conversations import Conversation
from rasa.llm_fine_tuning.paraphrasing.rephrased_user_message import (
    RephrasedUserMessage,
)
from rasa.shared.constants import (
    LLM_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    PROVIDER_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.providers.mappings import OPENAI_PROVIDER
from rasa.shared.utils.llm import (
    USER,
    get_prompt_template,
    llm_factory,
)

SEPARATOR = "\n\n"
BACKUP_SEPARATOR = "\nUSER:"

REPHRASING_PROMPT_FILE_NAME = "default_rephrase_prompt_template.jina2"
DEFAULT_REPHRASING_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.llm_fine_tuning.paraphrasing",
    REPHRASING_PROMPT_FILE_NAME,
)

DEFAULT_LLM_CONFIG = {
    PROVIDER_CONFIG_KEY: OPENAI_PROVIDER,
    MODEL_CONFIG_KEY: "gpt-4o-mini",
    TIMEOUT_CONFIG_KEY: 7,
    "temperature": 0.0,
    "max_tokens": 4096,
}

structlogger = structlog.get_logger()


class ConversationRephraser:
    def __init__(
        self,
        config: Dict[str, Any],
    ) -> None:
        self.config = {**self.get_default_config(), **config}
        self.prompt_template = get_prompt_template(
            self.config.get(PROMPT_TEMPLATE_CONFIG_KEY),
            DEFAULT_REPHRASING_PROMPT_TEMPLATE,
        )

    @staticmethod
    def validate_config(config: Dict[str, Any]) -> None:
        """Validate the rephrase_config."""
        if LLM_CONFIG_KEY in config:
            llm_config = config.get(LLM_CONFIG_KEY)

            # Check if LLM configuration is set to None, {}.
            if llm_config is None:
                error = "LLM config is empty. Please provide a valid LLM config."
                structlogger.error("rephrase_config.empty_llm_config", error=error)
                raise ValueError(error)

            # Validate LLM model name or model in config.
            if not llm_config.get(MODEL_CONFIG_KEY) and not llm_config.get(
                MODEL_NAME_CONFIG_KEY
            ):
                error = (
                    "LLM model name is empty. Please provide a valid LLM model name."
                )
                structlogger.error("rephrase_config.llm_model_is_not_set", error=error)
                raise ValueError(error)

        # Check if the config contains only the allowed keys.
        allowed_keys = {PROMPT_TEMPLATE_CONFIG_KEY, LLM_CONFIG_KEY}
        if len(set(config.keys()) - allowed_keys) > 0:
            error = (
                f"Invalid rephrase config. Only the following keys are allowed: "
                f"{', '.join(allowed_keys)}."
            )
            structlogger.error("rephrase_config.invalid_keys", error=error)
            raise ValueError(error)

    @staticmethod
    def get_default_config() -> Dict[str, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            PROMPT_TEMPLATE_CONFIG_KEY: None,
            LLM_CONFIG_KEY: DEFAULT_LLM_CONFIG,
        }

    async def rephrase_conversation(
        self, conversation: Conversation, number_of_rephrasings: int = 10
    ) -> List[RephrasedUserMessage]:
        """Create rephrasings for each user message in the conversation.

        For each user message create <number_of_rephrasings> number of rephrasings.
        The rephrasings are created with an LLM.

        Args:
            conversation: The conversation.
            number_of_rephrasings: The number of rephrasings to produce per user
            message.

        Returns:
            A list of rephrased user messages.
        """
        prompt = self._render_template(conversation, number_of_rephrasings)

        result = await self._invoke_llm(prompt)

        rephrased_user_messages = self._parse_output(
            result, conversation.get_user_messages()
        )

        # Check if user message was successfully rephrased
        self._check_rephrasings(
            rephrased_user_messages, number_of_rephrasings, result, conversation.name
        )

        return rephrased_user_messages

    async def _invoke_llm(self, prompt: str) -> str:
        llm = llm_factory(self.config.get(LLM_CONFIG_KEY), DEFAULT_LLM_CONFIG)
        try:
            llm_response = await llm.acompletion(prompt)
            return llm_response.choices[0]
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error(
                "conversation_rephraser.rephrase_conversation.llm.error", error=e
            )
            raise ProviderClientAPIException(e, message="LLM call exception")

    def _render_template(
        self,
        conversation: Conversation,
        number_of_rephrasings: int,
    ) -> str:
        user_messages = conversation.get_user_messages()
        number_of_user_messages = len(user_messages)

        return Template(self.prompt_template).render(
            test_case_name=conversation.name,
            transcript=conversation.transcript,
            number_of_user_messages=number_of_user_messages,
            number_of_rephrasings=number_of_rephrasings,
            user_prefix=USER,
            user_messages=user_messages,
        )

    @staticmethod
    def _extract_rephrasings(block: str) -> Tuple[Optional[str], Optional[List[str]]]:
        """Extract the rephrasings for a specific user message.

        Expected format looks like this:
        USER: <user message>
        1. <rephrased message 1>
        2. <rephrased message 2>
        3. <rephrased message 3>
        ...

        Args:
            block: String that contains the user message and its rephrasings.

        Returns:
            The original user message and the list of rephrasings.
        """
        if not block.strip():
            return None, None

        # Split the block by new line character
        lines = block.strip().split("\n")
        # Filter out empty lines and lines that are equal to """
        lines = [line.strip() for line in lines]
        lines = [line for line in lines if line and line != '"""']

        # We need at least the original user message and one rephrasing
        if len(lines) < 2:
            return None, None

        # Extract the user message from the first line
        # (ideally prefixed with 'USER: ')
        if lines[0].startswith("USER:"):
            original_user_message = lines[0][len(f"{USER}:") :].strip()
        else:
            original_user_message = lines[0]

        # Extract rephrasings
        rephrasings = []
        for line in lines[1:]:
            # Remove bullets or numbering and any extra whitespace
            line = re.sub(r"^\s*[-\d\.]+", "", line).strip()
            if line.startswith("USER:"):
                line = line[len(f"{USER}:") :].strip()
            if line:
                rephrasings.append(line)

        return original_user_message, rephrasings

    def _parse_output(
        self, output: str, user_messages: List[str]
    ) -> List[RephrasedUserMessage]:
        rephrased_messages = [
            RephrasedUserMessage(message, []) for message in user_messages
        ]

        message_blocks = self._get_message_blocks(output, len(rephrased_messages))
        for block in message_blocks:
            original_user_message, rephrasings = self._extract_rephrasings(block)

            if not original_user_message or not rephrasings:
                continue

            # Add the rephrasings to the correct user message
            for rephrased_message in rephrased_messages:
                if (
                    rephrased_message.original_user_message.lower()
                    == original_user_message.lower()
                ):
                    rephrased_message.rephrasings = rephrasings

        return rephrased_messages

    @staticmethod
    def _get_message_blocks(output: str, expected_number_of_blocks: int) -> List[str]:
        # Each user message block is (ideally) separated by new line ("\n\n")
        # USER: <user message>
        # 1. <rephrasing>
        #
        # USER: <user message>
        # 1. <rephrasing>
        message_blocks = output.split(SEPARATOR)
        if len(message_blocks) == expected_number_of_blocks:
            return message_blocks

        # In case the message blocks are not separated by new line, try to split it
        # by "\nUSER:", e.g.
        # USER: <user message>
        # 1. <rephrasing>
        # USER: <user message>
        # 1. <rephrasing>
        message_blocks = output.split(BACKUP_SEPARATOR)
        if len(message_blocks) == expected_number_of_blocks:
            return message_blocks

        return []

    @staticmethod
    def _check_rephrasings(
        rephrased_messages: List[RephrasedUserMessage],
        number_of_rephrasings: int,
        llm_output: str,
        conversation_name: str,
    ) -> None:
        incorrect_rephrasings_for_messages = []

        for message in rephrased_messages:
            if (
                not message.rephrasings
                or len(message.rephrasings) != number_of_rephrasings
            ):
                incorrect_rephrasings_for_messages.append(message.original_user_message)

        if incorrect_rephrasings_for_messages:
            structlogger.warning(
                "conversation_rephraser.rephrase_conversation.parse_llm_output",
                warning="Failed to parse llm output correctly. Not all user messages"
                "were successfully rephrased.",
                llm_output=llm_output,
                conversation_name=conversation_name,
                incorrect_rephrasings_for_messages=incorrect_rephrasings_for_messages,
            )



import logging
import typing
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text, Union

from faker import Faker

from rasa.shared.exceptions import RasaException
from rasa.utils.singleton import Singleton

if typing.TYPE_CHECKING:
    from presidio_analyzer import AnalyzerEngine
    from presidio_analyzer.nlp_engine.nlp_engine import NlpEngine
    from presidio_anonymizer.entities import OperatorConfig

DEFAULT_PRESIDIO_LANG_CODE = "en"
DEFAULT_PRESIDIO_MODEL_NAME = "en_core_web_lg"
DEFAULT_PRESIDIO_MODEL_PROVIDER = "spacy"

logger = logging.getLogger(__name__)


@dataclass
class AnonymizationRule:
    """A rule for anonymizing a given text."""

    entity_name: Text
    substitution: Text = "mask"
    value: Optional[Text] = None


@dataclass
class AnonymizationRuleList:
    """A list of anonymization rules."""

    id: Text
    rule_list: List[AnonymizationRule]
    language: Text = DEFAULT_PRESIDIO_LANG_CODE
    model_provider: Text = DEFAULT_PRESIDIO_MODEL_PROVIDER
    models: Union[Text, Dict[Text, Text]] = DEFAULT_PRESIDIO_MODEL_NAME


class AnonymizationAnalyzer(metaclass=Singleton):
    """Anonymization analyzer."""

    presidio_analyzer_engine = None

    def __init__(self, anonymization_rule_list: AnonymizationRuleList):
        """Initialise the anonymization analyzer."""
        if self.presidio_analyzer_engine is None:
            self.presidio_analyzer_engine = self._get_analyzer_engine(
                anonymization_rule_list
            )

    @staticmethod
    def _get_analyzer_engine(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> "AnalyzerEngine":
        """Returns an analyzer engine for all the anonymization rule lists."""
        from presidio_analyzer import AnalyzerEngine

        try:
            nlp_engine = AnonymizationAnalyzer._build_presidio_nlp_engine(
                anonymization_rule_list
            )
        except (OSError, ImportError) as exception:
            raise RasaException(
                "Failed to load Presidio nlp engine. "
                "Please check that you have provided "
                "a valid model name."
            ) from exception

        return AnalyzerEngine(
            nlp_engine=nlp_engine,
            supported_languages=[anonymization_rule_list.language],
        )

    @staticmethod
    def _build_presidio_nlp_engine(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> "NlpEngine":
        """Creates an instance of the Presidio nlp engine."""
        from presidio_analyzer.nlp_engine import (
            SpacyNlpEngine,
            StanzaNlpEngine,
            TransformersNlpEngine,
        )

        if anonymization_rule_list.model_provider == "transformers":
            nlp_engine = TransformersNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )
        elif anonymization_rule_list.model_provider == "stanza":
            nlp_engine = StanzaNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )
        else:
            nlp_engine = SpacyNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )

        return nlp_engine


class AnonymizationRuleExecutor:
    """Executes a given anonymization rule set on a given text."""

    def __init__(self, anonymization_rule_list: AnonymizationRuleList):
        """Initialize the anonymization rule executor."""
        from presidio_anonymizer import AnonymizerEngine

        self.anonymization_rule_list = anonymization_rule_list

        is_valid_rule_list = self._validate_anonymization_rule_list(
            anonymization_rule_list
        )

        self.analyzer = (
            AnonymizationAnalyzer(anonymization_rule_list)
            if is_valid_rule_list
            else None
        )

        self.anonymizer_engine = AnonymizerEngine()  # type: ignore

    @staticmethod
    def _validate_anonymization_rule_list(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> bool:
        """Validates the given anonymization rule list object."""
        if (
            anonymization_rule_list.language != DEFAULT_PRESIDIO_LANG_CODE
            and anonymization_rule_list.models == DEFAULT_PRESIDIO_MODEL_NAME
        ):
            logger.debug(
                f"Anonymization rule list language is "
                f"'{anonymization_rule_list.language}', "
                f"but no specific model name was provided. "
                f"You must specify the spaCy model name in the"
                f"endpoints yaml file. "
                f"Cannot proceed with anonymization."
            )
            return False

        return True

    def run(self, text: Text) -> Optional[Text]:
        """Anonymizes the given text using the given anonymization rule list."""
        if (
            self.analyzer is None
            or not self.anonymization_rule_list
            or self.analyzer.presidio_analyzer_engine is None
        ):
            return text

        if not self.anonymization_rule_list.rule_list:
            return text

        analyzer_results = self.analyzer.presidio_analyzer_engine.analyze(
            text=text,
            entities=[
                rule.entity_name for rule in self.anonymization_rule_list.rule_list
            ],
            language=self.anonymization_rule_list.language,
        )

        operators = self.get_operators()

        anonymized_text = self.anonymizer_engine.anonymize(
            text=text,
            analyzer_results=analyzer_results,
            operators=operators,
        )

        return anonymized_text.text

    @staticmethod
    @typing.no_type_check  # faker is not typed correctly
    def _get_supported_faker_entities() -> Dict[Text, Any]:
        faker = Faker(["en_US", "es_ES", "it_IT"])

        # Presidio entities: https://microsoft.github.io/presidio/supported_entities/
        # Faker providers: https://faker.readthedocs.io/en/master/providers.html
        # Unsupported entities by faker:
        # CRYPTO, NRP, MEDICAL_LICENSE, US_BANK_NUMBER, US_DRIVER_LICENSE
        # UK_NHS, IT_FISCAL_CODE, IT_DRIVER_LICENSE, IT_PASSPORT, IT_IDENTITY_CARD
        # SG_NRIC_FIN, AU_ABN, AU_ACN, AU_TFN, AU_MEDICARE
        supported_entities = {
            "PERSON": lambda value: faker["en_US"].name(),
            "PHONE_NUMBER": lambda value: faker["en_US"].phone_number(),
            "EMAIL_ADDRESS": lambda value: faker["en_US"].ascii_email(),
            "CREDIT_CARD": lambda value: faker["en_US"].credit_card_number(),
            "IBAN_CODE": lambda value: faker["en_US"].iban(),
            "DATE_TIME": lambda value: faker["en_US"].date(),
            "IP_ADDRESS": lambda value: faker["en_US"].ipv4(),
            "URL": lambda value: faker["en_US"].url(),
            # This faker method returns a tuple of
            # (latitude, longitude, place name, country code, timezone)
            "LOCATION": lambda value: faker["en_US"].location_on_land()[2],
            # USA
            "US_ITIN": lambda value: faker["en_US"].itin(),
            "US_PASSPORT": lambda value: faker["en_US"].passport_number(),
            "US_SSN": lambda value: faker["en_US"].ssn(),
            # Spain
            "ES_NIF": lambda value: faker["es_ES"].nif(),
            # Italy
            "IT_VAT_CODE": lambda value: faker["it_IT"].company_vat(),
        }

        return supported_entities

    @staticmethod
    def _mask_anonymize(value: Text) -> Text:
        return "*" * len(value)

    def get_substitution_func(self, rule: AnonymizationRule) -> Optional[Any]:
        """Returns a function that anonymizes the given text.

        Args:
            rule: The anonymization rule to use.

        Returns:
            A function that anonymizes the given text.
        """
        if rule.substitution == "faker":
            supported_faker_entities = self._get_supported_faker_entities()

            if rule.entity_name not in supported_faker_entities:
                logger.debug(
                    f"Unsupported faker entity: {rule.entity_name}. "
                    f"Supported entities are: {supported_faker_entities.keys()}"
                    f"Using mask anonymization instead."
                )
                func = self._mask_anonymize
            else:
                func = supported_faker_entities.get(rule.entity_name)
        elif rule.substitution == "mask":
            func = self._mask_anonymize
        else:
            raise RasaException(f"Unknown substitution type: {rule.substitution}")

        return func

    def get_operators(self) -> Dict[Text, "OperatorConfig"]:
        """Returns a dictionary of operators for the given anonymization rule list."""
        from presidio_anonymizer.entities import OperatorConfig

        operators = {}

        for rule in self.anonymization_rule_list.rule_list:
            if rule.substitution == "text":
                operators[rule.entity_name] = OperatorConfig(
                    "replace", {"new_value": rule.value}
                )
            else:
                operators[rule.entity_name] = OperatorConfig(
                    "custom", {"lambda": self.get_substitution_func(rule)}
                )

        return operators



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import structlog
from tqdm import tqdm

from rasa.llm_fine_tuning.conversations import Conversation, ConversationStep
from rasa.llm_fine_tuning.storage import StorageContext

LLM_DATA_PREPARATION_MODULE_STORAGE_LOCATION = "3_llm_finetune_data/llm_ft_data.jsonl"

structlogger = structlog.get_logger()


@dataclass
class LLMDataExample:
    prompt: str
    output: str
    original_test_name: str
    original_user_utterance: str
    rephrased_user_utterance: str

    def as_dict(self) -> Dict[str, Any]:
        return {
            "prompt": self.prompt,
            "output": self.output,
            "original_test_name": self.original_test_name,
            "original_user_utterance": self.original_user_utterance,
            "rephrased_user_utterance": self.rephrased_user_utterance,
        }


def _create_data_point(
    prompt: str,
    step: ConversationStep,
    conversation: Conversation,
    rephrased_user_message: Optional[str] = None,
) -> LLMDataExample:
    return LLMDataExample(
        prompt,
        step.commands_as_string(),
        conversation.get_full_name(),
        step.original_test_step.text,
        rephrased_user_message,
    )


def _update_prompt(
    prompt: str,
    original_user_steps: List[ConversationStep],
    rephrased_user_steps: List[str],
) -> Optional[str]:
    if len(original_user_steps) != len(rephrased_user_steps):
        structlogger.debug(
            "llm_fine_tuning.llm_data_preparation_module.failed_to_update_prompt",
            original_user_steps=[
                step.original_test_step.text for step in original_user_steps
            ],
            rephrased_user_steps=rephrased_user_steps,
        )
        return None

    updated_prompt = prompt
    for user_step, rephrased_message in zip(original_user_steps, rephrased_user_steps):
        # replace all occurrences of the original user message with the rephrased user
        # message in the conversation history mentioned in the prompt
        updated_prompt = updated_prompt.replace(
            f"USER: {user_step.original_test_step.text}", f"USER: {rephrased_message}"
        )

    # replace the latest user message mentioned in the prompt
    updated_prompt = updated_prompt.replace(
        f"'''{original_user_steps[-1].original_test_step.text}'''",
        f"'''{rephrased_user_steps[-1]}'''",
    )

    return updated_prompt


def _convert_conversation_into_llm_data(
    conversation: Conversation,
) -> List[LLMDataExample]:
    data = []

    # construct new conversations from the rephrasings
    new_conversations = _construct_new_conversations(conversation)

    original_user_steps = [
        step for step in conversation.iterate_over_annotated_user_steps()
    ]

    for i, step in enumerate(original_user_steps):
        # create data point for the original e2e test case
        data.append(_create_data_point(step.llm_prompt, step, conversation))

        # create data points using the rephrasings, e.g. 'new_conversations'
        for rephrased_user_steps in new_conversations:
            # +1 to include the current user turn
            prompt = _update_prompt(
                step.llm_prompt,
                original_user_steps[: i + 1],
                rephrased_user_steps[: i + 1],
            )
            if prompt:
                data.append(
                    _create_data_point(
                        prompt, step, conversation, rephrased_user_steps[i]
                    )
                )

    return data


def _construct_new_conversations(conversation: Conversation) -> List[List[str]]:
    """Construct new conversations from the rephrasings.

    In general, we will combine the passing rephrasings at the same index position to
    construct a new conversation. If for one particular user turn no other passing
    rephrasing exists, we reset the index and take the first passing rephrasing again.

    Args:
        conversation: The conversation.

    Returns:
        A list of new conversations (only rephrased user turns).
    """
    max_passed_rephrasings = max(
        [
            len(step.passed_rephrasings)
            for step in conversation.iterate_over_annotated_user_steps()
        ]
    )

    new_conversations = []
    for i in range(0, max_passed_rephrasings):
        current_conversation = []
        for step in conversation.iterate_over_annotated_user_steps():
            # take the orginial user message in case no passing rephrasings exist
            if not step.passed_rephrasings and step.original_test_step.text:
                structlogger.debug(
                    "llm_fine_tuning.llm_data_preparation_module."
                    "construct_new_conversations.no_passed_rephrasings",
                    conversation=conversation.get_full_name(),
                    step=step.original_test_step.text,
                    message="Take original user message instead of rephrasing.",
                )
                current_conversation.append(step.original_test_step.text)
                continue

            # some user steps might have less rephrasings than others
            # loop over the rephrasings
            index = i % len(step.passed_rephrasings)
            current_conversation.append(step.passed_rephrasings[index])
        if current_conversation:
            new_conversations.append(current_conversation)

    structlogger.debug(
        "llm_fine_tuning.llm_data_preparation_module.construct_new_conversations",
        conversation=conversation.get_full_name(),
        new_conversations=new_conversations,
    )

    return new_conversations


def convert_to_fine_tuning_data(
    conversations: List[Conversation], storage_context: StorageContext
) -> List[LLMDataExample]:
    llm_data = []

    for i in tqdm(range(len(conversations))):
        llm_data.extend(_convert_conversation_into_llm_data(conversations[i]))

    storage_context.write_llm_data(
        llm_data, LLM_DATA_PREPARATION_MODULE_STORAGE_LOCATION
    )

    return llm_data



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

from rasa.dialogue_understanding.commands.command import Command
from rasa.dialogue_understanding.patterns.user_silence import (
    UserSilencePatternFlowStackFrame,
)
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class UserSilenceCommand(Command):
    """A command to indicate user silence."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "user silence"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> UserSilenceCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return UserSilenceCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        stack.push(UserSilencePatternFlowStackFrame())
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(self.command())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, UserSilenceCommand)
